(* Boxed version of the main interfaces in VPL hierarchy *)

Require Export NumC.
Require Export ProgVar.
Require Export Debugging.
Require Import Itv.
Require Export ImpureConfig.

Create HintDb vpl discriminated.

Ltac VPLTsimplify :=
  xtsimplify ltac:(eauto with vpl).

Ltac VPLAsimplify :=
  xasimplify ltac:(eauto with vpl).

(* TODO: cut BasicDomain in smaller pieces ? *)

Module Type BasicDomain (N: NumSig).

  Parameter t: Type.

  Parameter gamma: t -> Mem.t N.t -> Prop.

  Declare Instance gamma_ext: Proper (Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> iff) gamma.

  Parameter isIncl: t -> t -> imp bool.
  Parameter isIncl_correct: forall a1 a2, If (isIncl a1 a2) THEN forall m, gamma a1 m -> gamma a2 m.

  Parameter top: t.
  Parameter top_correct: forall m, gamma top m.

  Parameter join: t -> t -> imp t.
  Parameter join_correct: forall a1 a2, WHEN p <- join a1 a2 THEN forall m, gamma a1 m \/ gamma a2 m -> gamma p m.

  Parameter widen: t -> t -> imp t.

  Parameter bottom: t.
  Parameter bottom_correct: forall m, (gamma bottom m)->False.

  Parameter project: t -> PVar.t -> imp t.
  Parameter project_correct: forall a x, WHEN p <- project a x THEN forall m v, gamma a m -> gamma p (Mem.assign x v m).

  Hint Resolve isIncl_correct top_correct join_correct bottom_correct project_correct: vpl.

End BasicDomain.


Module Type HasIsBottom(N:NumSig) (Import D: BasicDomain N).
  
  Parameter isBottom: t -> imp bool.

  Parameter isBottom_correct: forall a, If isBottom a THEN forall m, (gamma a m)->False.

  Hint Resolve isBottom_correct: vpl. 

End HasIsBottom.

(** A generic test for contradiction *)
Module AddIsBottom (N: NumSig) (Import D: BasicDomain N) <: HasIsBottom N D.
  
  Definition isBottom (a:t) : imp bool :=
    isIncl a bottom.

  Lemma isBottom_correct a: If isBottom a THEN forall m, (gamma a m)->False.
  Proof.
    VPLAsimplify.
  Qed.

  Global Opaque isBottom.

  Hint Resolve isBottom_correct: vpl. 

End AddIsBottom.


Module Type HasSimplePrinting (N: NumSig) (Import D: BasicDomain N).
 
  Parameter pr: t -> String.string.

End HasSimplePrinting.


Module Type HasRename (N: NumSig) (Import D: BasicDomain N).
  Require Import Program.Basics.

  (** [rename x y a] renames variable [x] by [y] in [a].

   WARNING: For the underlying "shadow" implementation in ML, 
              [y] must be a fresh variable in [a]
            However, we do not need to make this precondition explicit 
            for correctness (thus, this is simpler to let it hidden).

    Hence, this precondition is only dynamically verified by an "assert" in ML code.
  *)
  Parameter rename: PVar.t -> PVar.t -> t -> imp t.
  Parameter rename_correct: forall x y a, WHEN p <- (rename x y a) THEN
    forall m, gamma a (Mem.assign x (m y) m) -> gamma p m.
    (* ie. forall x y a, isRename gamma impl (rename x y a) a x y. *)


End HasRename.

(** Interval from Terms *)

Module Type TermSig(N:NumSig).
  
  Parameter t: Type.
  Parameter eval: t -> Mem.t N.t -> N.t.

End TermSig.

Module Type HasGetItv (N: NumSig) (Import T: TermSig N) (Import I: ItvSig N) (Import D: BasicDomain N).

  Parameter get_itv: T.t -> t -> imp I.t.

  Parameter get_itv_correct: forall (te: T.t) (a: t), WHEN i <- (get_itv te a) THEN
    forall m, gamma a m -> I.sat i (T.eval te m).

  Hint Resolve get_itv_correct: vpl.

End HasGetItv.

(** Assume/Require of conditions  *)

Module Type CondSig(N:NumSig).
  
  Parameter t: Type.
  Parameter sat: t -> Mem.t N.t -> Prop.

End CondSig.


Module Type HasAssume (N: NumSig) (Cond: CondSig N) (D:BasicDomain N).

  Import Cond.
  Import D.

  Parameter assume: Cond.t -> t -> imp t.
  Parameter assume_correct: forall c a, WHEN p <- assume c a THEN forall m, sat c m -> gamma a m -> gamma p m.

  Hint Resolve assume_correct: vpl.

End HasAssume.


Module Type HasRequire (N: NumSig) (Cond: CondSig N) (D:BasicDomain N).
  
  Import Cond.
  Import D.

  Parameter require: Cond.t -> t -> imp bool.
  Parameter require_correct: forall c a, If require c a THEN forall m, gamma a m -> sat c m.

  Hint Resolve require_correct: vpl.

End HasRequire.


Module Type HasConcreteConditions (N: NumSig) (Cond: CondSig N) (D: BasicDomain N) 
  := HasAssume N Cond D <+ HasRequire N Cond D.

Module Type WeakAbstractDomain (N: NumSig) (Cond: CondSig N) 
  := BasicDomain N <+ HasIsBottom N <+ HasAssume N Cond.

Module Type AbstractDomain (N: NumSig) (Cond: CondSig N) 
  := WeakAbstractDomain N Cond <+ HasRequire N Cond.


(* Assignement *)
Module Type XCondSig(N:NumSig).
  
  Declare Module Term: TermSig N.
  Include CondSig N.

  (** two state semantics:
    [xsat c old new] means that "c" is a relation between "old" state and "new" state.
  *)
  Parameter xsat: t -> Mem.t N.t -> Mem.t N.t -> Prop.

  Parameter xsat_sat: forall c m, xsat c m m = sat c m.

End XCondSig.


Module Type HasAssign (N: NumSig) (Cond: XCondSig N) (D: BasicDomain N).
  
  Import Cond.Term.
  Import Cond.
  Import D.
  
  (* deterministic assignement *)
  Parameter assign: PVar.t -> Term.t -> t -> imp t.
  Parameter assign_correct: forall x te a, WHEN p <- assign x te a THEN
    forall m, gamma a m -> gamma p (Mem.assign x (eval te m) m).

  (* guarded assignement *)
  Parameter guassign: PVar.t -> Cond.t -> t -> imp t.
  Parameter guassign_correct: forall x c a, WHEN p <- guassign x c a THEN 
     forall m v, xsat c m (Mem.assign x v m) -> gamma a m -> gamma p (Mem.assign x v m).

End HasAssign.


Module Type FullAbstractDomain (N: NumSig) (Cond: XCondSig N) := AbstractDomain N Cond <+ HasAssign N Cond.
Module Type FullItvAbstractDomain (N: NumSig) (Cond: XCondSig N) (I: ItvSig N) := FullAbstractDomain N Cond <+ HasGetItv N Cond.Term I.
