(** A Demo: a Hoare-triple verifier for a tiny imperative programming language called PL 

This little demo illustrates:
 - how to use our polyhedral abstract domain on Z, called "FullPolLinZ".
 - how to program and reason on impure computations.

*)

Require Import BoxedPolLinZ.
Require Import AffCond.
Import ZCond.
Import Term.

Local Open Scope impure.

Local Notation var := positive.
Local Notation value := Z.
Local Notation mem := (Mem.t ZNum.t).

(** * SPECIFICATION PART *)

(** ** Abstract syntax of statement of PL *)
Inductive statement: Type :=
 | Assign (x: var) (t: term)
 | Seq (s1 s2: statement)
 | Ifs (c: cond) (s1 s2: statement)
 | While (c: cond) (s: statement) (inv: cond)
  .

(** Notations (concrete syntax) *)
Bind Scope statement_scope with statement.
Bind Scope cond_scope with cond.
Bind Scope term_scope with term.
Delimit Scope statement_scope with statement.
Delimit Scope cond_scope with cond.
Delimit Scope term_scope with term.

Definition minus (x y: term) := Plus x (Opp y).

Infix "+" := Plus: term_scope.
Infix "-" := minus: term_scope.
Infix "*" := Mult: term_scope.
Infix "<=" := (Atom Le): cond_scope.
Infix "<" := (Atom Lt): cond_scope.
Infix "=" := (Atom Eq): cond_scope.
Infix "<>" := (Atom Neq): cond_scope.
Infix "/\" := (BinL AND): cond_scope.
Infix "\/" := (BinL OR): cond_scope.

Infix "::=" := Assign
  (at level 54, no associativity): statement_scope.

Notation "s1 -;  s2" := (Seq s1 s2): statement_scope.

Notation "'If' c 'Then' s1 'Else' s2 'Fi'" :=(Ifs c s1 s2)
  (at level 95, c at level 95, no associativity):  statement_scope.
Notation "'While' c 'Do' s 'Inv' pi 'Done'" :=(While c s pi)
  (at level 95, c at level 95, no associativity):  statement_scope.

(** ** Big-steps semantics of statement *)
Open Scope statement.

Inductive sem: statement -> mem -> mem -> Prop  :=
 | sem_assign x t m: 
     sem (x ::= t) m (Mem.assign x (Term.eval t m) m)
 | sem_seq s1 s2 m0 m1 m2:
     sem s1 m0 m1 
     -> sem s2 m1 m2 
     -> sem (s1 -; s2) m0 m2
 | sem_ifs_true c s1 s2 m0 m1: 
     sat c m0 
     -> sem s1 m0 m1
     -> sem (If c Then s1 Else s2 Fi) m0 m1
 | sem_ifs_false c s1 s2 m0 m1: 
     ~(sat c m0) 
     -> sem s2 m0 m1
     -> sem (If c Then s1 Else s2 Fi) m0 m1
 | sem_while_true c s m0 m1 inv:
    sat c m0
    -> sem (s -; While c Do s Inv inv Done) m0 m1
    -> sem (While c Do s Inv inv Done) m0 m1
 | sem_while_false c s inv m: 
     ~(sat c m) 
     -> sem (While c Do s Inv inv Done) m m
 .

Record hoareTriple : Type :=
 { pre: cond ;
   prog: statement ;
   post: cond }.

Definition partially_correct (h:hoareTriple) : Prop :=
  forall m m', sat (pre h) m -> sem (prog h) m m' -> sat (post h) m'. 

(** * IMPLEMENTATION PART *)

(** An equivalent semantics, with a simpler inductive principle on while_true case *)

Inductive semX: statement -> mem -> mem -> Prop  :=
 | semX_assign x t m: 
     semX (x ::= t) m (Mem.assign x (Term.eval t m) m)
 | semX_seq s1 s2 m0 m1 m2:
     semX s1 m0 m1 
     -> semX s2 m1 m2 
     -> semX (s1 -; s2) m0 m2
 | semX_ifs_true c s1 s2 m0 m1: 
     sat c m0 
     -> semX s1 m0 m1
     -> semX (If c Then s1 Else s2 Fi) m0 m1
 | semX_ifs_false c s1 s2 m0 m1: 
     ~(sat c m0) 
     -> semX s2 m0 m1
     -> semX (If c Then s1 Else s2 Fi) m0 m1
 | semX_while_false c s inv m: 
     ~(sat c m) 
     -> semX (While c Do s Inv inv Done) m m
 | semX_while_true c s m0 m1 m2 inv:
    sat c m0
    -> semX s m0 m1
     -> (semX (While c Do s Inv inv Done) m1 m2)
      -> semX (While c Do s Inv inv Done) m0 m2
 .

Hint Resolve semX_assign semX_seq semX_ifs_true semX_ifs_false semX_while_false semX_while_true: sem.


Lemma sem_semX s m0 m1: sem s m0 m1 -> semX s m0 m1. 
Proof.
  induction 1; eauto with sem.
  (* while_true *)
  inversion IHsem; eauto with sem.
Qed.
Hint Resolve sem_semX: demo.


(** ** An alternate semantics by wlp *)

Fixpoint wlp (s: statement) (post: mem -> Prop) (m: mem) : Prop :=
 match s with
 | (x ::= t) => post (Mem.assign x (Term.eval t m) m)
 | (s1 -; s2) => wlp s1 (wlp s2 post) m
 | (If c Then s1 Else s2 Fi) => ((sat c m) -> (wlp s1 post m)) /\  (~(sat c m) -> (wlp s2 post m))
 | (While c Do s Inv _ Done) => 
    exists inv: mem->Prop, 
     (forall m', sat c m' -> inv m' -> wlp s inv m')
     /\ (inv m) 
     /\ (forall m', ~(sat c m') -> inv m' -> post m')
 end.
 
Lemma wlp_monotonic s: forall (P Q: mem -> Prop) m,
   wlp s P m 
   ->  (forall m, P m -> Q m) 
   -> wlp s Q m.
Proof.
  induction s; simpl; intuition simpl; eauto.
  firstorder.
Qed.

Hint Resolve wlp_monotonic: demo.

Lemma wlp_correctness s m m': (semX s m m') -> forall post, wlp s post m -> post m'.
Proof.
  induction 1; simpl in * |- *; intuition eauto.
  (* seq *) 
   - firstorder.
  (* if_true *)
   - firstorder.
  (* if_false *)
   - firstorder.
  (* while_false *)
   - firstorder.
  (* while_true *)
   - eapply IHsemX2; eauto; clear IHsemX2.
     decompose [ex and] H2; clear H2.
     eapply ex_intro; firstorder.
     eapply IHsemX1; eauto.
Qed.

Lemma wlp_True s:forall m, wlp s (fun m' => True) m.
Proof.
  induction s; simpl; intuition eauto with demo.
Qed.

(** Auxiliary computations on polyhedra *)
Definition assume c p := FullPolLinZ.assume (nnfForAssume c) p.

Lemma assume_correct c a:  
 WHEN p <- assume c a THEN
   forall m, sat c m -> FullPolLinZ.gamma a m -> FullPolLinZ.gamma p m.
Proof.
  VPLAsimplify.
Qed.
Hint Resolve assume_correct: demo.
Global Opaque assume.

Definition require c p := FullPolLinZ.require (nnfForRequire c) p.

Lemma require_correct c a: 
  WHEN b <- require c a THEN
    b=true -> forall m, FullPolLinZ.gamma a m -> sat c m.
Proof.
  VPLAsimplify.
  intros; subst; simpl in * |-. eauto with vpl.
Qed.
Hint Resolve require_correct: demo.
Global Opaque require.

Definition embed c : imp FullPolLinZ.t :=
  assume c FullPolLinZ.top.

(** Computations of polyhedral postcondition *)

Fixpoint postfinder pre s: imp FullPolLinZ.t :=
  match s with
  | (x ::= t) => 
      FullPolLinZ.assign x t pre
  | (s1 -; s2) =>
      BIND aux <- postfinder pre s1 -;
      postfinder aux s2
  | (If c Then s1 Else s2 Fi) =>
      BIND pre1 <- assume c pre -;
      BIND post1 <- postfinder pre1 s1 -;
      BIND pre2 <- assume (Not c) pre -;
      BIND post2 <- postfinder pre2 s2 -;
      FullPolLinZ.join post1 post2
  | (While c Do s Inv inv Done) =>
      BIND pi <- embed inv -;
      BIND b <- FullPolLinZ.isIncl pre pi -;
      if b 
      then
         BIND pre1 <- assume c pi -;
         BIND post <- postfinder pre1 s -;
         BIND b' <- FullPolLinZ.isIncl post pi -;
         if b'
         then assume (Not c) pi
         else failwith "invariant preservation" (embed (Not c))
       else
         failwith "invariant init" (embed (Not c))
  end.

Lemma postfinder_correct s: forall pre, 
  WHEN post <- postfinder pre s THEN 
      forall m, FullPolLinZ.gamma pre m -> wlp s (FullPolLinZ.gamma post) m.
Proof.
  induction s; simpl; unfold failwith, embed; xasimplify ltac:(eauto with vpl demo).
  (* ifs *)
  - simpl in * |- *.
    intros X1 m X2; intuition; eapply wlp_monotonic. 
    + eapply H0; eauto.
    + eauto.
    + eapply H2; eauto.
    + eauto.
  (* while normal case *)
  - simpl in * |- *.
    intros X1 m X2.
    constructor 1 with (x:=FullPolLinZ.gamma exta).
    intuition eauto with demo.
  (* while error case 1 *)
  - simpl in * |- *.
    intros X1 m X2.
    constructor 1 with (x:=fun m' => True).
    intuition.
    apply wlp_True.
  (* while error case 2 *)
  - simpl in * |- *.
    intros X1 m X2.
    constructor 1 with (x:=fun m' => True).
    intuition.
    apply wlp_True.
Qed.
Hint Resolve postfinder_correct: demo.

(** * MAIN THEOREM *)

Definition hoareTripleVerifier (h: hoareTriple) : imp bool :=
   BIND init <- embed (pre h) -;
   BIND p <- postfinder init (prog h) -;
   require (post h) p.

(** This theorem ensures that when "hoareTripleVerifier h" computes "true"
    then "h" is correct (in partial correctness) !
*) 
Theorem hoareTripleVerifier_correct (h: hoareTriple):
  WHEN b <- hoareTripleVerifier h THEN 
      b=true -> partially_correct h.
Proof.
  unfold partially_correct, hoareTripleVerifier, embed; destruct h; simpl.
  xasimplify ltac:(eauto with demo vpl).
  simpl; intros; eapply wlp_correctness; eauto with vpl demo.
Qed.
