(** Impure monad for interface with impure code

Little reasoning support is provided in ltac (similar to OptionMonad).

*)

Require Export Setoid.
Require Export OptionMonad.

Module Type ImpureMonad.

  Axiom imp: Type -> Type.
  Axiom pure: forall {A}, A -> imp A.
  Axiom bind: forall {A B}, imp A -> (A -> imp B) -> imp B.

  Axiom mayReturn: forall {A:Type}, imp A -> A -> Prop.

  Axiom mayReturn_pure: forall (A:Type) (a b:A), 
     mayReturn (pure a) b -> a=b.

  Axiom mayReturn_bind: forall (A B:Type) (k1:imp A) (k2: A -> imp B) (b:B),
     mayReturn (bind k1 k2) b -> exists a:A, mayReturn k1 a /\ mayReturn (k2 a) b.

  Axiom impeq: forall {A}, imp A -> imp A -> Prop.

  Declare Instance impeq_equiv A: Equivalence (@impeq A).
  Declare Instance bind_eq_compat A B: Proper ((@impeq A) ==> (pointwise_relation A (@impeq B)) ==> (@impeq B)) (@bind A B).
  Declare Instance mayReturn_eq_compat A: Proper ((@impeq A) ==> Logic.eq ==> iff) (@mayReturn A).

  Axiom impeq_bind_pure_l: forall (A B:Type) (k: A -> imp B) (a:A),
     (impeq (bind (pure a) k) (k a)).

  Axiom impeq_bind_pure_r: forall (A:Type) (k: imp A),
     (impeq (bind k (fun b => pure b)) k).

  Axiom impeq_bind_assoc: forall (A B C:Type) (k1: imp A) (k2: A -> imp B) (k3: B -> imp C),
     (impeq (bind (bind k1 k2) (fun b => k3 b)) (bind k1 (fun a => bind (k2 a) (fun b => k3 b)))).


End ImpureMonad.

Module Type TYPE.
   Axiom t: Type.
End TYPE.

(** model of state-monad computations *)
Module StateImpureMonad(State:TYPE): ImpureMonad.
   
   Definition imp (A:Type) := State.t -> A * State.t.
   Definition pure {A:Type} (a:A) := fun (s:State.t) => (a,s).
   Definition bind {A B:Type} (k1: imp A) (k2: A -> imp B) := 
     fun s0 => let r := k1 s0 in k2 (fst r) (snd r).

   Definition mayReturn {A:Type} (k: imp A) (a:A): Prop := 
     exists s, fst (k s)=a.

   Lemma mayReturn_pure (A:Type) (a b:A): (mayReturn (pure a) b) -> a=b.
   Proof.
     unfold mayReturn, pure. simpl.
     firstorder. 
   Qed.

   Lemma mayReturn_bind (A B:Type) (k1:imp A) (k2: A -> imp B) (b:B):
         (mayReturn (bind k1 k2) b) -> exists (a:A), (mayReturn k1 a) /\ (mayReturn (k2 a) b).
   Proof.
     unfold mayReturn, bind. 
     firstorder eauto.
   Qed.

  Definition impeq {A} (k1 k2: imp A) := forall s, (k1 s)=(k2 s).

  Instance impeq_equiv A: Equivalence (@impeq A).
  Proof.
    unfold impeq; constructor 1; intro x; congruence.
  Qed.

  Instance bind_eq_compat A B: Proper ((@impeq A) ==> (pointwise_relation A (@impeq B)) ==> (@impeq B)) (@bind A B).
  Proof.
    unfold impeq, bind; intros x y; congruence.
  Qed.

  Instance mayReturn_eq_compat A: Proper ((@impeq A) ==> Logic.eq ==> iff) (@mayReturn A).
  Proof.
    unfold impeq, mayReturn, iff; intros k1 k2 H x y H0.
    subst; firstorder subst;
    eapply ex_intro; rewrite H; eauto.
  Qed.

  Lemma impeq_bind_pure_l (A B:Type) (k: A -> imp B) (a:A):
     (impeq (bind (pure a) k) (k a)).
  Proof.
    unfold impeq, pure, bind; simpl; auto.
  Qed.

  Lemma impeq_bind_pure_r (A:Type) (k: imp A):
     (impeq (bind k (fun b => pure b)) k).
  Proof.
    unfold impeq, pure, bind. intros s; destruct (k s); simpl; auto.
  Qed.

  Lemma impeq_bind_assoc (A B C:Type) (k1: imp A) (k2: A -> imp B) (k3: B -> imp C):
     (impeq (bind (bind k1 k2) (fun b => k3 b)) (bind k1 (fun a => bind (k2 a) (fun b => k3 b)))).
  Proof.
    unfold impeq, bind. intros s; simpl; auto.
  Qed.

End StateImpureMonad.

(** model of potentially non-terminating, imperative, non-deterministic computations *)
Module StandardImpureMonad(State:TYPE): ImpureMonad.
 
   Definition imp (A:Type) := State.t -> State.t -> A -> Prop.
   Definition pure {A:Type} (a:A) := fun (s0 s1:State.t) b => s0=s1 /\ a=b.
   Definition bind {A B:Type} (k1: imp A) (k2: A -> imp B) := 
     fun s0 s2 b => exists (s1:State.t), exists (a:A), k1 s0 s1 a /\ k2 a s1 s2 b.

   Definition mayReturn {A:Type} (k: imp A) (a:A): Prop := exists s0, exists s1, k s0 s1 a.

   Lemma mayReturn_pure {A:Type} (a b:A): (mayReturn (pure a) b) -> a=b.
   Proof.
     unfold mayReturn, pure. firstorder.
   Qed.

   Lemma mayReturn_bind {A B:Type} (k1:imp A) (k2: A -> imp B) (b:B):
         (mayReturn (bind k1 k2) b) -> exists (a:A), (mayReturn k1 a) /\ (mayReturn (k2 a) b).
   Proof.
     unfold mayReturn, bind. 
     firstorder.
   Qed.

  Definition impeq {A} (k1 k2: imp A) := forall s0 s1 a, (k1 s0 s1 a) <-> (k2 s0 s1 a).

  Instance impeq_equiv A: Equivalence (@impeq A).
  Proof.
    unfold impeq; constructor 1; intro x.
    - intuition.
    - firstorder.
    - intros y z H H0 s0 s1 a. rewrite H. auto.
  Qed.

  Instance bind_eq_compat A B: Proper ((@impeq A) ==> (pointwise_relation A (@impeq B)) ==> (@impeq B)) (@bind A B).
  Proof.
    unfold impeq, bind; intros x y H; unfold pointwise_relation; simpl.
    intros f g H0 s0 s1 a. constructor 1;
    intros H1; decompose [ex and] H1; clear H1.
    rewrite H, H0 in * |-; firstorder.
    rewrite <- H, <- H0 in * |-; firstorder.
  Qed.

  Instance mayReturn_eq_compat A: Proper ((@impeq A) ==> Logic.eq ==> iff) (@mayReturn A).
  Proof.
    unfold impeq, mayReturn; intros k1 k2 H x y H0.
    subst; firstorder subst.
    rewrite H in * |-; firstorder.
    rewrite <- H in * |-; firstorder.
  Qed.

  Lemma impeq_bind_pure_l (A B:Type) (k: A -> imp B) (a:A):
     (impeq (bind (pure a) k) (k a)).
  Proof.
    unfold impeq, pure, bind; firstorder subst; auto.
  Qed.

  Lemma impeq_bind_pure_r (A:Type) (k: imp A):
     (impeq (bind k (fun b => pure b)) k).
  Proof.
    unfold impeq, pure, bind; firstorder subst; auto.
  Qed.

  Lemma impeq_bind_assoc (A B C:Type) (k1: imp A) (k2: A -> imp B) (k3: B -> imp C):
     (impeq (bind (bind k1 k2) (fun b => k3 b)) (bind k1 (fun a => bind (k2 a) (fun b => k3 b)))).
  Proof.
    unfold impeq, bind;  firstorder subst;
    repeat (eapply ex_intro; intuition eauto).
  Qed.

End StandardImpureMonad.


Module ImpureMonadTheory(Export M: ImpureMonad).

  Definition impeq_bind_pure_l:=impeq_bind_pure_l.
  Definition impeq_bind_pure_r:=impeq_bind_pure_r.
  Definition impeq_bind_assoc:=impeq_bind_assoc.

  Hint Rewrite @impeq_bind_pure_l @impeq_bind_pure_r @impeq_bind_assoc: impeq.
  Hint Immediate (fun A x => @Equivalence_Reflexive _ _ (@impeq_equiv A) x).

  Definition wlp {A:Type} (k: imp A) (P: A -> Prop): Prop
    := forall a, mayReturn k a -> P a.


(* compatibility with rewriting *)

  Add Parametric Morphism (A:Type): (@wlp A) with
    signature (@impeq A) ==> (pointwise_relation A iff) ==> iff
    as wlp_compat.
  Proof.
    intros k1 k2 H P Q H0. unfold pointwise_relation, iff in H.
    constructor 1; unfold wlp; intros.
    rewrite <- H in * |- ; firstorder.
    rewrite H in * |-; firstorder.
  Qed.

(* wlp lemmas for tactics *)

  Lemma wlp_unfold A (k:imp A)(P: A -> Prop):
    (forall a, mayReturn k a -> P a)
    -> wlp k P.
  Proof.
    auto.
  Qed.

  Lemma wlp_monotone A (k:imp A) (P1 P2: A -> Prop):
    wlp k P1
    -> (forall a, mayReturn k a -> P1 a -> P2 a)
    -> wlp k P2.
  Proof.
    unfold wlp; eauto.
  Qed.

  Lemma wlp_forall A B (k:imp A) (P: B -> A -> Prop):
    (forall x, wlp k (P x))
    -> wlp k (fun a => forall x, P x a).
  Proof.
    unfold wlp; auto.
  Qed.

  Lemma wlp_pure A (P: A -> Prop) a:
    P a -> wlp (pure a) P.
  Proof.
    unfold wlp.
    intros H b H0; 
      rewrite <- (mayReturn_pure _ _ _ H0).
    auto.
  Qed.

  Lemma wlp_bind: forall A B (k1:imp A) (k2: A -> imp B) (P: B -> Prop),
    wlp k1 (fun a => wlp (k2 a) P) -> wlp (bind k1 k2) P.
  Proof.
    unfold wlp.
    intros A B k1 k2 P H a H0.
    case (mayReturn_bind _ _ _ _ _ H0); clear H0.
    intuition eauto.
  Qed.


(** Example of Higher-order reasoning *)
  Lemma List_fold_left_impeq_run (A B:Type) (f: B -> A -> imp A) (l:list B) (i: imp A):
   impeq (List.fold_left (fun k x => bind k (f x)) l i)
         (bind i (fun a => List.fold_left (fun k x => bind k (f x)) l (pure a))).
  Proof.
    simpl. 
    generalize i.
    induction l as [| a0 l IHl]; simpl.
    - intros; autorewrite with impeq; auto.
    - intros; rewrite IHl.
      autorewrite with impeq.
      apply bind_eq_compat. 
      + auto.
      + intros a. rewrite IHl; auto;
           autorewrite with impeq; auto.  
  Qed.

  Lemma wlp_List_fold_left (A B:Type) (f: B -> A -> imp A) (l:list B) (i: imp A) (P: A -> Prop): 
    wlp i (fun a => wlp (List.fold_left (fun k x => bind k (f x)) l (pure a)) P)
     -> wlp (List.fold_left (fun k x => bind k (f x)) l i) P.
  Proof.
    intros H; rewrite List_fold_left_impeq_run.
    apply wlp_bind. auto.
  Qed.

(* Tactics 

MAIN tactics:  
  - xtsimplify "base": simplification using from hints in "base" database (in particular "wlp" lemmas).
  - xtstep "base": only one step of simplification.

For good performance, it is recommanded to have several databases.

*)

  Ltac introcomp :=
    let a:= fresh "exta" in
      let H:= fresh "Hexta" in
        intros a H.

(* decompose the current wlp goal using "introduction" rules *)
  Ltac wlp_decompose :=
    apply wlp_pure
 || apply wlp_bind.

(* this tactic simplifies the current "wlp" goal using any hint found via tactic "hint". *) 
  Ltac apply_wlp_hint hint :=
    eapply wlp_monotone;
      [ hint; fail | idtac ] ;
      simpl; introcomp.

(* one step of xsimplify 
see xtsimplify below
*)
  Ltac xtstep hint :=
    match goal with
      | |- (wlp _ _) => wlp_decompose
        || apply_wlp_hint ltac:hint
          || (apply wlp_unfold; 
            introcomp)
    end.

(* main general tactic 
WARNING: for the good behavior of "xtsimplify", "hint" must at least perform a "eauto".

Example of use:
  xtsimplify ltac:(intuition eauto with base).
*)
  Ltac xtsimplify hint := 
    repeat (intros; xtstep ltac:hint ; simpl; (tauto || hint)).

  Ltac xastep hint :=
    match goal with
      | |- (wlp _ _) => wlp_decompose
        || apply_wlp_hint ltac:hint
      | |- (wte _ _ _) => wte_decompose
        || apply_wte_hint ltac:hint
          || (apply wte_decomp; [ intro | idtac]; intro_rewrite)
      | |- (ifprop _ _ _) => ifprop_decompose
        || apply_ifprop_hint ltac:hint
          || (apply ifprop_decomp; intro_rewrite)
      | |- _ =>    default_simplify
        || (apply wlp_unfold; 
          introcomp)
    end.


  Ltac xasimplify hint:=
    repeat (intros; xastep ltac:hint; simpl; (tauto || hint)).


(* Notations *)

  Bind Scope impure_scope with imp.
  Delimit Scope impure_scope with impure.
  Open Scope impure.

  Notation "'BIND' x '<-' k1 '-;' k2" := (bind k1 (fun x => k2))
    (at level 55, k1 at level 53, x at level 99, right associativity): impure_scope.

  Notation "'SOME' x '<-' k1 '-;' k2" := (TRY x <- k1 IN k2 CATCH (pure None))
    (at level 55, k1 at level 53, right associativity): impure_scope.

  Notation "k1 '-;'  k2" := (bind k1 (fun x:unit => k2))
    (at level 55, right associativity): impure_scope.

  Notation "'WHEN' a '<-' k 'THEN' R" := (wlp k (fun a => R))
    (at level 73, R at level 95, right associativity): impure_scope.

  Notation "'If' k 'THEN' R" := (wlp k (fun b => ifprop b R True))
    (at level 73, R at level 95, right associativity): impure_scope.
End ImpureMonadTheory.
