Require String.
Require Export OptionMonad.
Require Export NumC.
Require Export ProgVar.
Require Export Debugging.
Require Import PositiveMapAddOn.
Require Import Sorting.Permutation.


(* database used until "PolC" construction *)
Create HintDb polC discriminated.

(* a tiny wrapper to OptionMonad.simplify *)
Ltac PolCsimplify :=
  xsimplify ltac:(eauto with polC).


Module Type LinSig.
(** The interface of that an implementation of linear terms must provide. *)

  Parameter t: Type.

  Parameter eval: t -> Mem.t QNum.t -> QNum.t.

  Parameter mayDependOn: t -> PVar.t -> Prop.
 
  Parameter eval_mdo: mdoExt mayDependOn eval Logic.eq.

  Parameter nil: t.
  Parameter NilEval: forall m, eval nil m = QNum.z.

  Parameter single: PVar.t -> QNum.t -> t.
  Parameter single_correct: forall x n m, eval (single x n) m = QNum.mul (m x) n.

  Parameter neg: t -> t.
  Parameter NegsatSpec: forall lt m, eval (neg lt) m = QNum.neg (eval lt m).

  Parameter mul: QNum.t -> t -> t.
  Parameter MulsatSpec: forall n lt m, eval (mul n lt) m = QNum.mul n (eval lt m).

  Definition Eq (lt1 lt2: t): Prop
    := forall m, eval lt1 m = eval lt2 m.

  Parameter isEq: t -> t -> bool.
  Parameter IsEqsatSpec: forall lt1 lt2, If isEq lt1 lt2 THEN Eq lt1 lt2.

  Parameter add: t -> t -> t.
  Parameter AddsatSpec: forall lt1 lt2 m, eval (add lt1 lt2) m = QNum.add (eval lt1 m) (eval lt2 m).


  Parameter isFree: PVar.t -> t -> bool.
  Parameter isFree_correct: forall x l, If isFree x l THEN ~(mayDependOn l x).
  Hint Resolve IsEqsatSpec isFree_correct:polC.
  Hint Rewrite NilEval AddsatSpec MulsatSpec NegsatSpec single_correct: linterm.

  (** Rename of a variable by an arbitrary variable. 
      we choose a strong version because proofs are simpler in CstrC.v and ConsSet.v !
  *)
  Parameter rename: PVar.t -> PVar.t -> t -> t.
  Parameter rename_correct: forall (x y:PVar.t) (l:t) m, 
    (eval (rename x y l) m)=(eval l (Mem.assign x (m y) m)).

  Parameter pr: t -> String.string.

End LinSig.

Module PositiveMapVec <: LinSig.
(** Implementation of linear terms as radix map-tree from variable to coefficients (PositiveMap.t).

Implicit invariant for this data structure: there is no explicit zero coefficient, nor un-needed 
"None" in the PositiveMap (or in other word, PositiveMaps are only handled in normal form).

However, this invariant is not needed for correctness (but only for efficiency & precision).
Hence, we do not prove it !
*)

  Local Open Scope type_scope.
  Local Open Scope list_scope.
  Import List.ListNotations.

  Add Ring NumRing: QNum.Ring.

  Definition t: Type
    := PositiveMap.t QNum.t.

  Definition absEval (l: list (PVar.t*QNum.t)) (m:Mem.t QNum.t) : QNum.t :=
   List.fold_right (fun p sum => QNum.add sum (QNum.mul (m (fst p)) (snd p)))
                       QNum.z  l.

  Definition eval (lt:t) (m:Mem.t QNum.t) : QNum.t
    := absEval (PositiveMap.elements lt) m.

  Definition mayDependOn (lt:t) (x: PVar.t) : Prop 
    := List.fold_right (fun p sum => x=(fst p) \/ sum)
                        False
                        (PositiveMap.elements lt).

  Lemma mayDependOn_altDef (lt:t) (x: PVar.t): mayDependOn lt x <-> exists c, List.In (x,c) (PositiveMap.elements lt).
  Proof.
    unfold mayDependOn; induction (PositiveMap.elements lt); simpl; firstorder (subst; eauto).
    destruct a; simpl. eapply ex_intro; eauto.
  Qed.

  Lemma eval_mdo: mdoExt mayDependOn eval Logic.eq.
  Proof.
    unfold mdoExt, bExt, mayDependOn, eval. intro e; induction (PositiveMap.elements e); simpl; auto.
  Qed. 
 
  Definition nil: t
    := PositiveMap.Leaf.

  Lemma NilEval: forall m, eval nil m = QNum.z.
  Proof.
    intro m;
    reflexivity.
  Qed.

  Definition single (x:PVar.t) (n:QNum.t) : t :=
    if QNum.eqDec n QNum.z then nil else single x n.

  Lemma single_correct: forall x n m, eval (single x n) m = QNum.mul (m x) n.
  Proof.
    intros; unfold single.
    case (QNum.eqDec n QNum.z).
    - intros; subst; ring_simplify; auto.
    - intros; unfold eval; rewrite elements_single; simpl.
      ring.
  Qed.

  Definition neg (lt:t) := PositiveMap.map (fun c => QNum.neg c) lt.

  Lemma NegsatSpec: forall lt m, eval (neg lt) m = QNum.neg (eval lt m).
  Proof.
    intros; unfold eval, neg; rewrite elements_map.
    generalize (PositiveMap.elements lt); induction l; simpl; auto.
    rewrite IHl. clear IHl; ring_simplify. auto. 
  Qed.

  Definition mul (n:QNum.t) (lt:t) : t
    := if QNum.eqDec n QNum.z 
       then nil
       else PositiveMap.map (fun c => QNum.mul n c) lt.

  Lemma MulsatSpec: forall n lt m, eval (mul n lt) m = QNum.mul n (eval lt m).
  Proof.
    intros; unfold mul.
    case (QNum.eqDec n QNum.z).
    - intros; subst; ring_simplify; auto.
    - intros; unfold eval; rewrite elements_map.
      generalize (PositiveMap.elements lt); induction l; simpl.
      ring.
      rewrite IHl. clear IHl; ring_simplify. auto.
  Qed. 
 
  Definition Eq (lt1 lt2: t): Prop
    := forall m, eval lt1 m = eval lt2 m.

  Definition QNum_eqb (x y: QNum.t): bool :=QNum.eqDec x y ||| false.

  Lemma QNum_eqb_correct: forall x y, If QNum_eqb x y THEN x=y.
  Proof.
    unfold QNum_eqb; intros; PolCsimplify.
  Qed.

  Definition isEq: t -> t -> bool := (equal QNum_eqb).
  
  Local Hint Resolve (equal_correct _ QNum_eqb_correct) : PolC.
  Lemma IsEqsatSpec: forall lt1 lt2, If isEq lt1 lt2 THEN Eq lt1 lt2.
  Proof.
    unfold isEq. xsimplify ltac:(eauto with PolC).
    intros; subst; unfold Eq; auto.
  Qed.

  Definition add: t -> t -> t
    := merge (fun n1 n2 => let r:=QNum.add n1 n2 in
                            if QNum.eqDec r QNum.z then None else Some r).

  Lemma AddsatSpec: forall lt1 lt2 m, eval (add lt1 lt2) m = QNum.add (eval lt1 m) (eval lt2 m).
  Proof.
    unfold add, eval; intros lt1 lt2.
    set (f:=(fun n1 n2 => let r:=QNum.add n1 n2 in
                            if QNum.eqDec r QNum.z then None else Some r)).
    generalize (elements_merge f lt1 lt2).
    generalize (PositiveMap.elements lt1) (PositiveMap.elements lt2) (PositiveMap.elements (merge f lt1 lt2)).
    unfold f; clear f.
    induction 1; simpl in * |- *.
    - intros; ring_simplify; auto.
    - intros; ring_simplify; auto.
    - intros; rewrite IHinMerge. ring_simplify; auto.
    - intros; rewrite IHinMerge. 
      destruct (QNum.eqDec (QNum.add v1 v2) QNum.z) as [X | X]; try discriminate.
      injection H; intro; subst. ring_simplify; auto.
    - intros; rewrite IHinMerge. destruct (QNum.eqDec (QNum.add v1 v2) QNum.z) as [X|X]; try discriminate.
      cutrewrite (v2 = QNum.sub QNum.z v1).
      ring_simplify; auto.
      rewrite <- X; ring.
    - intros; rewrite IHinMerge. ring_simplify; auto.
  Qed.
  Hint Rewrite NilEval AddsatSpec MulsatSpec NegsatSpec single_correct: linterm.


  Definition isFree (x: PVar.t) (l: t): bool := negb (PositiveMap.mem x l).

  Lemma isFree_correct: forall x l, If (isFree x l) THEN ~(mayDependOn l x).
  Proof.
    unfold isFree; PolCsimplify.
    rewrite mayDependOn_altDef.
    intro H; destruct H as [v H].
    erewrite In_mem in Heq_simplified; eauto.
    discriminate.
  Qed.
  Hint Resolve isFree_correct IsEqsatSpec: polC.

  Lemma absEval_permutation l l' m: Permutation l l' -> absEval l m=absEval l' m.
  Proof.
    induction 1; simpl; congruence || ring.
  Qed.

  Lemma eval_remove_free x (l:t) v m: eval (PositiveMap.remove x l) (Mem.assign x v m)= eval (PositiveMap.remove x l) m.
  Proof.
    unfold eval. rewrite elements_remove.
    induction (PositiveMap.elements l); simpl; auto.
    unfold eq_key at 1 4.
    destruct a as (x1,c1); simpl.
    destruct (BinPos.Pos.eq_dec x x1); simpl; auto.
    rewrite Mem.assign_out; auto.
  Qed.

  (** Below, a rename that works even if [y] is not free in [l] 
      It is simple to prove using previous lemmas.

      Actually, exploiting the fact that [y] is free, we could optimize
      a bit using:
          [PositiveMap.add y c (PositiveMap.remove x l)]
      instead of [add (single y c) (PositiveMap.remove x l)]

      However, this tiny optimization would results in much harder proofs in CstrC.v
      and ConsSet.v ! 
  *)
  Definition rename (x y: PVar.t) (l:t) : t :=
    match PositiveMap.find x l with
    | Some c => add (single y c) (PositiveMap.remove x l)
    | None => l
    end.

  Lemma rename_correct (x y:PVar.t) (l:t) m: 
    (eval (rename x y l) m)=(eval l (Mem.assign x (m y) m)).
  Proof.
    unfold rename. 
    generalize (find_remove_elements_Permutation x l) (eval_remove_free x l (m y) m).
    case (PositiveMap.find x l); simpl.
    - intros v H. autorewrite with linterm. 
      unfold eval. rewrite <- (absEval_permutation _ _ _ H).
      simpl. autorewrite with progvar.
      intros H1; rewrite H1; ring.
    - unfold eval. intros H; rewrite H; auto.
  Qed.

  Import String.
  Local Open Scope string_scope.

  Fixpoint fmtAux (l: list string): string -> string
    := fun s =>
         match l with
           | [] => s
           | prem::l' => fmtAux l' (prem ++ s)
         end.

  Definition fmt (l: list string): string
    :=  match l with
           | [] => "0"
           | prem::l' => fmtAux l' prem
         end.

  Definition pairPr (p: PVar.t * QNum.t) : string
    := (PVar.pr (fst p)) ++ " * " ++ (QNum.pr (snd p)).

  Definition pr: t -> string
    := fun lt => 
           let l:=(PositiveMap.elements lt) in fmt (List.rev (List.map pairPr l)).

  Hint Rewrite NilEval AddsatSpec MulsatSpec NegsatSpec single_correct: linterm.

End PositiveMapVec.

Module Lin <: LinSig.

  Include PositiveMapVec.

  Definition exportT: Set
    := list (PVar.t * QNum.t).

  Definition export: t -> exportT
    := fun lt => (PositiveMap.elements lt).

  Definition import: exportT -> t
    := fun l => List.fold_left (fun lt p => add (single (fst p) (snd p)) lt) l nil.

End Lin.
