(** This module gives an abstract syntax (inspired from logic formula) for polyhedra 
    
    Actually, the syntax is a bit more expressive than declared above, 
    since "conditions" are arbitrary boolean formula where atoms are 
    affine constrains.

    At, the end of the module, some functors are provided to lift domains 
    with a "assume/require" on Cstr into "assume/require" on these conditions.

*)

Require Import Bool.
Require Import ZArith.
Require Export AffTerm.
Require Import LinTerm.
Require Import CstrC.
Require Import OptionMonad.
Require Export BoxedDomainInterfaces.

Inductive cmpT: Set := Eq | Le | Lt | Neq.

Inductive binl: Set := AND | OR.

Definition binlDenote (og: binl) : Prop -> Prop -> Prop
  := if og then and else or.
Extraction Inline binlDenote.  


(* polyhedra as logical expressions *)
Module Cond (N: NumSig) (Import Term: AffTermSig(N)). 

  Definition cmpDenote: cmpT -> N.t -> N.t -> Prop
    := fun cmp x1 x2 =>
      match cmp with
        | Eq => x1 = x2
        | Le => N.Le x1 x2
        | Lt => N.Lt x1 x2
        | Neq => x1 <> x2
      end.

  Lemma cmpDenote_dec (cmp: cmpT) (n1 n2: N.t): {cmpDenote cmp n1 n2} + {~ cmpDenote cmp n1 n2}. 
    refine (match cmp return {cmpDenote cmp n1 n2} + {~ cmpDenote cmp n1 n2} with
              | Eq => N.eqDec n1 n2
              | Le =>
                match N.ltLeDec n2 n1 with
                  | left pf => right (proj1 (N.LtNotLe n2 n1) pf)
                  | right pf => left pf
                end
              | Lt =>
                match N.ltLeDec n1 n2 with
                  | left pf => left pf
                  | right pf => right (proj1 (N.LeNotLt n2 n1) pf)
                end
              | Neq =>
                match N.eqDec n1 n2 with
                  | left pf => right (fun H => H pf)
                  | right pf => left pf
                end
            end).
  Qed.
  Hint Resolve cmpDenote_dec: vpl.

  Inductive cond: Set :=
  | Basic (b: bool)
  | Atom (oc: cmpT) (tl tr: term)
  | BinL (op: binl) (cl cr: cond)
  | Not (c:cond) .

  Coercion Basic: bool >-> cond.
  Coercion Atom: cmpT >-> Funclass.
  Coercion BinL: binl >-> Funclass.

  Definition t := cond.

    (* Semantics *)
  Fixpoint sat (c:t) (m: Mem.t N.t): Prop := 
    match c with
      | Basic b => Is_true b
      | Atom oc tl tr => cmpDenote oc (eval tl m) (eval tr m)
      | BinL op cl cr => binlDenote op (sat cl m) (sat cr m)
      | Not c => ~(sat c m)
    end. 

  Lemma sat_dec (c: t) (m: Mem.t N.t): {sat c m} + {~ sat c m}.
  Proof.
    induction c; simpl; auto with vpl.
    - case b; simpl; auto.
    - case op; simpl; intuition.
    - intuition.
  Qed.

  Fixpoint xsat (c:t) (old new: Mem.t N.t): Prop := 
    match c with
      | Basic b => Is_true b
      | Atom oc tl tr => cmpDenote oc (xeval tl old new) (xeval tr old new)
      | BinL op cl cr => binlDenote op (xsat cl old new) (xsat cr old new)
      | Not c => ~(xsat c old new)
    end. 

  Lemma xsat_dec (c: t) (old new: Mem.t N.t): {xsat c old new} + {~ xsat c old new}.
  Proof.
    induction c; simpl; auto with vpl.
    - case b; simpl; auto.
    - case op; simpl; intuition.
    - intuition.
  Qed.

  (* Useless: derived from eval_pointwise_compat ?  
    Add Parametric Morphism: sat with
    signature Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> Logic.eq
    as sat_compat.
    Proof.
       intros c m1 m2 H; induction c; simpl; try congruence.
       case oc; simpl; rewrite H; auto.
    Qed.
   *)

  Lemma xsat_sat c m: xsat c m m = sat c m.
  Proof.
    induction c; simpl; autorewrite with vpl; try congruence.
  Qed.
  Hint Rewrite xsat_sat: vpl.

  (* Useless: derived from eval_pointwise_compat ?
    Add Parametric Morphism: xsat with
    signature Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> Logic.eq
    as xsat_compat.
    Proof.
     intros c o1 o2 H1 n1 n2 H2; induction c; simpl; try congruence.
     case oc; simpl; rewrite H1; rewrite H2; auto.
    Qed.
  *)

   (** mayDependOn for conditions *)
  Fixpoint mayDependOn (c: cond) (x: PVar.t): Prop := 
    match c with
      | Basic _ => False
      | Atom _ tl tr => Term.mayDependOn tl x \/ Term.mayDependOn tr x 
      | BinL _ cl cr => mayDependOn cl x \/ mayDependOn cr x
      | Not c => mayDependOn c x
    end.

  Hint Resolve (@mdBounds_fold _ mayDependOn): progvar.

  Fixpoint mdBound (c:cond) (bound: positive): positive := 
    match c with
      | Basic _ => bound
      | Atom _ tl tr => Term.mdBound tl (Term.mdBound tr bound) 
      | BinL _ cl cr => mdBound cl (mdBound cr bound)
      | Not c => mdBound c bound
    end.

  Program Instance mdBound_mdBoundVar: MDBoundVar (mayDependOn:=mayDependOn) (mdBound:=mdBound).
  Obligation 1.
    generalize bound; induction e; simpl; intros; autorewrite with rwmax; auto.
    rewrite Term.mdBound_mdBoundVar_obligation_1 with (e:=tl);
    rewrite Term.mdBound_mdBoundVar_obligation_1 with (e:=tr);
    apply eq_sym.
    rewrite Term.mdBound_mdBoundVar_obligation_1.
    autorewrite with rwmax; auto.
    rewrite IHe1; rewrite IHe2. 
    apply eq_sym.
    rewrite IHe1; autorewrite with rwmax; auto.     
  Qed.
  Hint Resolve mdBound_mdBoundVar_obligation_1: progvar.
  Obligation 2.
    unfold mdBounds; generalize bound; induction e; simpl; intros; subst; intuition; eauto with progvar.
  Qed.
  Hint Resolve mdBound_mdBoundVar_obligation_2: progvar.

  Lemma sat_mdo: mdoExt mayDependOn sat Logic.eq.
  Proof.
    unfold mdoExt, bExt; induction e; simpl; intuition.
    - erewrite Term.eval_mdo with (m2:=m2); eauto. 
    erewrite Term.eval_mdo with (e:=tr) (m2:=m2); intuition eauto.
    - erewrite IHe; eauto.
  Qed.

  Lemma xsat_old_mdo old: mdoExt mayDependOn (fun c => xsat c old) Logic.eq.
  Proof.
    unfold mdoExt, bExt; induction e; simpl; intuition.
    - erewrite Term.xeval_old_mdo; auto.
    erewrite Term.xeval_old_mdo with (e:=tr); eauto.
    - erewrite IHe; eauto.
  Qed.

  Lemma xsat_new_mdo new: mdoExt mayDependOn (fun c old => xsat c old new) Logic.eq.
  Proof.
    unfold mdoExt, bExt; induction e; simpl; intuition.
    - erewrite xeval_new_mdo; auto.
    erewrite xeval_new_mdo with (e:=tr); eauto.
    - erewrite IHe; eauto.
  Qed.

  Fixpoint map (c:cond) f : cond :=
    match c with
      | Atom oc tl tr => Atom oc (Term.map tl f) (Term.map tr f) 
      | BinL op cl cr => BinL op (map cl f) (map cr f)
      | Not c => Not (map c f)
      | c0 => c0
    end.

  Lemma sat_map: forall c f m, sat (map c f) m = sat c (fun x => m (f x)).
  Proof.
    induction c; simpl; auto with vpl.
  Qed.

  Fixpoint xmap (c:cond) old new : cond := 
    match c with
      | Atom oc tl tr => Atom oc (Term.xmap tl old new) (Term.xmap tr old new) 
      | BinL op cl cr => BinL op (xmap cl old new) (xmap cr old new)
      | Not c => Not (xmap c old new)
      | c0 => c0
    end.

  Lemma xsat_xmap: forall c old new f g, xsat (xmap c f g) old new = xsat c (fun x => old (f x)) (fun x => new (g x)).
  Proof.
    induction c; simpl; auto with vpl.
  Qed.
  Hint Resolve xsat_xmap: vpl.

  Section Nnf.
  (* 

  Elimination of negation in conditions (Negative Normal Form) 

  This elimination is parametrized by a "reformulation" of Lt and Neq.

  Two reasons:
   - On Z domains, the linear constrain "t1 < t2" generated in Q arithmetic  
     is not very sharp. Hence it is more precise to encode it using "(t2+1) <= t1"
   - Abstraction of "Neq" as a OR can leads to a loss of precision on a "require"
     Alternatively, "require (t1 <> t2)" is better implemented by "assume (t1=t2); require false".

  Hence: we may want to remove "Neq" for assume (using a "OR" for instance), but keep it for "require" !

  The soundness (resp. completeness) hypotheses on these alternate encodings is only used
  in the corresponding lemma of Nnf. The alternate formulations must be themselves in "NNF".
  This is not "formally" required, because we do not need to prove that the resulting condition 
  has no negation (if this is false, this will only result in a loss of precision).

  *)

    Variable altLt: term -> term -> t. 

    Hypothesis altLt_sound: forall t1 t2 old new, (xsat (Lt t1 t2) old new) -> (xsat (altLt t1 t2) old new).
    Hypothesis altLt_complete: forall t1 t2 old new, (xsat (altLt t1 t2) old new) -> (xsat (Lt t1 t2) old new).

    Section Nnf_parametrized_by_Neq.

      Variable altNeq: term -> term -> t.
      Hypothesis altNeq_sound: forall t1 t2 old new, (xsat (Neq t1 t2) old new) -> (xsat (altNeq t1 t2) old new).

      Definition alt (cmp: cmpT) (t1 t2: term): t :=
        match cmp with
          | Lt => altLt t1 t2
          | Neq => altNeq t1 t2
          | cmp => cmp t1 t2
        end.
      Extraction Inline alt.

      Lemma alt_sound: forall (cmp:cmpT) t1 t2 old new, (xsat (cmp t1 t2) old new) -> xsat (alt cmp t1 t2) old new.
      Proof.
        intros cmp t1 t2 old new; case cmp; simpl; auto.
      Qed.
      Hint Resolve alt_sound: vpl.

      Definition dual (cmp: cmpT) (t1 t2: term): t :=
        match cmp with
          | Eq => altNeq t1 t2
          | Le => altLt t2 t1
          | Lt => Le t2 t1
          | Neq => Eq t1 t2
        end.
      Extraction Inline dual.

      Lemma dual_sound: forall (cmp:cmpT) t1 t2 old new, ~(xsat (cmp t1 t2) old new) -> xsat (dual cmp t1 t2) old new.
      Proof.
        intros cmp t1 t2 old new; case cmp; simpl; autorewrite with num; auto.
        intro H; case (N.eqDec (xeval t2 old new) (xeval t1 old new)); auto.
        intros; case H; auto.
      Qed.
      Hint Resolve dual_sound: vpl.

      Fixpoint nnf (c:t):  t :=
        match c with
          | Atom op tl tr => alt op tl tr
          | BinL op cl cr => op (nnf cl) (nnf cr)
          | Not c0 => nnfNot c0
          | c0 => c0
        end
      with nnfNot (c:t):  t :=
        match c with
          | Basic b => negb b
          | Atom op tl tr => dual op tl tr
          | BinL AND cl cr => OR (nnfNot cl) (nnfNot cr)
          | BinL OR cl cr => AND (nnfNot cl) (nnfNot cr)
          | Not c0 => nnf c0
        end.

      Fixpoint nnf_xsound (c:t) old new {struct c}: (xsat c old new) -> (xsat (nnf c) old new)
      with nnfNot_xsound (c:t) old new {struct c}: ~(xsat c old new) -> (xsat (nnfNot c) old new).
      Proof.
        * destruct c; simpl; auto with vpl.
        case op; simpl; intuition.
        * destruct c; simpl; auto with vpl.
        - case b; simpl; auto.
        - case op; simpl; intuition.
        case (xsat_dec c1 old new); intuition.
        - intros H; case (xsat_dec c old new); intuition.
      Qed.
      Hint Resolve nnf_xsound: vpl.

      Lemma nnf_sound (c:cond) m: (sat c m) -> (sat (nnf c) m).
      Proof.
        rewrite <- !xsat_sat. auto with vpl.
      Qed.
      Hint Resolve nnf_sound: vpl.


      Hypothesis altNeq_complete: forall t1 t2 old new, (xsat (altNeq t1 t2) old new) -> (xsat (Neq t1 t2) old new).

      Lemma alt_complete: forall (cmp:cmpT) t1 t2 old new, xsat (alt cmp t1 t2) old new -> cmpDenote cmp (xeval t1 old new) (xeval t2 old new).
      Proof.
        intros cmp t1 t2 old new; case cmp; simpl in * |- *; auto.
      Qed.
      Hint Resolve alt_complete: vpl.

      Lemma dual_complete: forall (cmp:cmpT) t1 t2 old new, xsat (dual cmp t1 t2) old new -> ~(cmpDenote cmp (xeval t1 old new) (xeval t2 old new)).
      Proof.
        intros cmp t1 t2 old new; case cmp; simpl in * |- *; autorewrite with num; auto.
      Qed.
      Hint Resolve dual_complete: vpl.

      Fixpoint nnf_xcomplete (c:t) old new {struct c}: (xsat (nnf c) old new) -> (xsat c old new)
      with NnfNot_xcomplete (c:t) old new {struct c}: (xsat (nnfNot c) old new) -> ~(xsat c old new).
      Proof.
        * destruct c; simpl; auto with vpl.
        case op; simpl; intuition.
        * destruct c; simpl; auto with vpl.
        - case b; simpl; auto.
        - case op; simpl; intuition eauto.
      Qed.
      Hint Resolve nnf_xcomplete: vpl.

      Lemma nnf_complete (c:cond) m: (sat (nnf c) m) -> (sat c m).
      Proof.
        rewrite <- !xsat_sat. auto with vpl.
      Qed.
      Hint Resolve nnf_complete: vpl.

      (* Nnf and mayDependOn *)
      Hypothesis altLt_mdBound: forall t1 t2 x, mdBound (altLt t1 t2) x = Term.mdBound t1 (Term.mdBound t2 x).
      Hypothesis negLt_mdBound: forall t1 t2 x, mdBound (altNeq t1 t2) x = Term.mdBound t1 (Term.mdBound t2 x).
       
      Fixpoint nnf_mdBound (c:t) x {struct c}: mdBound (nnf c) x = mdBound c x
      with nnfNot_mdBound (c:t) x {struct c}: mdBound (nnfNot c) x = mdBound c x.
      Proof.
        * destruct c; simpl; auto.
           - case oc; simpl; auto.
           - rewrite !nnf_mdBound; auto.
        * destruct c; simpl; auto.
           - case oc; simpl; auto;
             try (rewrite altLt_mdBound; auto);
             erewrite mdBound_comm; auto.
           - case op; simpl;
             rewrite !nnfNot_mdBound; auto.
      Qed.
    End Nnf_parametrized_by_Neq.

    (* Default implementation of Neq as a OR *)
    Definition altNeqByOR t1 t2 := OR (altLt t1 t2) (altLt t2 t1).

    Lemma altNeqByOR_sound: forall t1 t2 old new, (xsat (Neq t1 t2) old new) -> (xsat (altNeqByOR t1 t2) old new).
    Proof.
      unfold altNeqByOR. simpl.
      intros t1 t2 old new; case (N.dec (xeval t1 old new) (xeval t2 old new)); intuition.
    Qed.

    Lemma nLt_diff n1 n2: n1=n2 -> N.Lt n1 n2 -> False.
    Proof.
      intro; subst. rewrite N.LtNotLe; auto with num.
    Qed.
    Hint Resolve nLt_diff: num.

    Lemma altNeqByOR_complete: forall t1 t2 old new, (xsat (altNeqByOR t1 t2) old new) -> (xsat (Neq t1 t2) old new).
    Proof.
      unfold altNeqByOR. simpl in * |- *; intuition eauto with num.
    Qed.

  End Nnf.
  
  Extraction Inline altNeqByOR.

  Lemma altNeqByOR_stdLt_sound: forall t1 t2 old new, (xsat (Neq t1 t2) old new) -> (xsat (altNeqByOR (Atom Lt) t1 t2) old new).
  Proof.
    intros; eapply altNeqByOR_sound; eauto.
  Qed.

  Lemma altNeqByOR_stdLt_complete: forall t1 t2 old new, (xsat (altNeqByOR (Atom Lt) t1 t2) old new) -> (xsat (Neq t1 t2) old new).
  Proof.
    intros. eapply altNeqByOR_complete; eauto.
  Qed.

  Hint Resolve nnf_xsound nnf_sound altNeqByOR_stdLt_sound nnf_complete altNeqByOR_stdLt_complete: affcond.

  (*
  Extraction nnf.
  *)

End Cond.

Module Type AffCondSig (N: NumSig). (* <: XCondSig N with Module Term := Term. *)

  Declare Module Term: AffTermSig N.

  Include Cond N Term.

  (* See encoding of assume and require *)

  Parameter nnfForAssume : t -> t.
  Parameter nnfForAssume_xsound: forall c old new, (xsat c old new) -> (xsat (nnfForAssume c) old new).
  Parameter nnfForAssume_sound: forall c m, (sat c m) -> (sat (nnfForAssume c) m).

  Parameter nnfForRequire : t -> t.
  Parameter nnfForRequire_complete: forall c m, (sat (nnfForRequire c) m) -> (sat c m).

  Hint Resolve nnfForAssume_sound: vpl.

End AffCondSig.



Module ZCond <: AffCondSig ZNum.

  Module Term := ZTerm.

  Include Cond ZNum ZTerm.

  Import ZTerm.
   (* Local Open Scope Z_scope. *)

  Definition altLt t1 t2 := Le (Plus t1 (Cte 1%Z)) t2.
  Extraction Inline altLt.

  Lemma altLt_sound: forall t1 t2 old new, (xsat (Lt t1 t2) old new) -> (xsat (altLt t1 t2) old new).
  Proof.
    unfold altLt, trace; simpl; unfold ZNum.Lt, ZNum.Le, ZNum.add. intros. omega.
  Qed. 
  Hint Resolve altLt_sound: affcond.

  Lemma altLt_complete: forall t1 t2 old new, (xsat (altLt t1 t2) old new) -> (xsat (Lt t1 t2) old new).
  Proof.
    unfold altLt, trace; simpl; unfold ZNum.Lt, ZNum.Le, ZNum.add. intros. omega.
  Qed. 
  Hint Resolve altLt_complete: affcond.

  Lemma altNeqByOR_altLt_sound: forall t1 t2 old new, (xsat (Neq t1 t2) old new) -> (xsat (altNeqByOR altLt t1 t2) old new).
  Proof.
    intros; eapply altNeqByOR_sound; eauto with affcond.
  Qed.

  Lemma altNeqByOR_altLt_complete: forall t1 t2 old new, (xsat (altNeqByOR altLt t1 t2) old new) -> (xsat (Neq t1 t2) old new).
  Proof.
    intros. eapply altNeqByOR_complete. 2:eauto.
    auto with affcond.
  Qed.
  Hint Resolve altNeqByOR_altLt_sound altNeqByOR_altLt_complete: affcond.


  Definition nnfForAssume := 
    (ZCond.nnf ZCond.altLt (ZCond.altNeqByOR ZCond.altLt)).

  Lemma nnfForAssume_sound c m: (sat c m) -> (sat (nnfForAssume c) m).
  Proof.
    unfold nnfForAssume, trace; eauto with affcond.
  Qed.
  Hint Resolve nnfForAssume_sound: vpl.

  Lemma nnfForAssume_xsound: forall c old new, (xsat c old new) -> (xsat (nnfForAssume c) old new).
  Proof.
    unfold nnfForAssume, trace; eauto with affcond.
  Qed.
  Hint Resolve nnfForAssume_xsound: vpl.

  Lemma nnfForAssume_mdBound: forall c x, ZCond.mdBound (nnfForAssume c) x = ZCond.mdBound c x.
  Proof.
    unfold nnfForAssume; apply nnf_mdBound; simpl; auto.
    intros; rewrite mdBound_invol. rewrite mdBound_comm with (e1:=t2).
    rewrite mdBound_invol. auto.
  Qed.
  Global Opaque nnfForAssume.

  (* Using Lt instead of altLt should be better on require ? *)
  Definition nnfForRequire := 
    (ZCond.nnf (Atom Lt) (* altLt *)
              (ZCond.Atom Neq)).

  Lemma nnfForRequire_complete: forall c m, (sat (nnfForRequire c) m) -> (sat c m).
  Proof.
    unfold nnfForRequire, trace; eauto with affcond.
  Qed.
  Hint Resolve nnfForRequire_complete: vpl.
  Global Opaque nnfForRequire.


End ZCond.


Module QCond <: AffCondSig QNum.

  Module Term := QTerm.  

  Include Cond QNum QTerm.

  Definition nnfForAssume :=
    QCond.nnf (QCond.Atom Lt) 
    (QCond.altNeqByOR (QCond.Atom Lt)).

  Lemma nnfForAssume_sound c m: (sat c m) -> (sat (nnfForAssume c) m).
  Proof.
    unfold nnfForAssume; eauto with affcond.
  Qed.
  Hint Resolve nnfForAssume_sound: vpl.

  Lemma nnfForAssume_xsound: forall c old new, (xsat c old new) -> (xsat (nnfForAssume c) old new).
  Proof.
    unfold nnfForAssume; eauto with affcond.
  Qed.
  Hint Resolve nnfForAssume_xsound: vpl.

  Definition nnfForRequire :=
    QCond.nnf (Atom Lt) (QCond.Atom Neq).

  Lemma nnfForRequire_complete: forall c m, (sat (nnfForRequire c) m) -> (sat c m).
  Proof.
    unfold nnfForRequire; eauto with affcond.
  Qed.
  Hint Resolve nnfForRequire_complete: vpl.
  

  (* translation from ZConds to QConds *)
  Lemma lift_cmp cmp n1 n2: cmpDenote cmp (ZtoQ.ofZ n1) (ZtoQ.ofZ n2) <-> ZCond.cmpDenote cmp n1 n2.
  Proof.
    case cmp; simpl; autorewrite with num; intuition.
  Qed.
  Hint Rewrite lift_cmp: vpl.

  Fixpoint lift (c: ZCond.t): cond :=
    match c with
      | ZCond.Basic b => Basic b
      | ZCond.Atom oc tl tr => Atom oc (QTerm.lift tl) (QTerm.lift tr)
      | ZCond.BinL op cl cr => BinL op (lift cl) (lift cr)
      | ZCond.Not c => Not (lift c)
    end.

  Lemma lift_correct (c: ZCond.t) (m: Mem.t ZNum.t):
    sat (lift c) (Mem.lift ZtoQ.ofZ m) <-> ZCond.sat c m.  
  Proof.
    induction c; simpl; autorewrite with vpl; intuition;
      destruct op; simpl in * |- *; intuition.
  Qed.
  Hint Rewrite lift_correct: vpl.


  (** Translations from AST into Cstr.t *)
  Add Ring NumRing: QNum.Ring.  

  Definition toCstrX (t1: QTerm.t) (cmp:CstrC.cmpT) (t2:QTerm.t) : Cstr.t :=
    let (lt1,aux1) := QTerm.toLin t1 in
      let (lt2,aux2) := QTerm.toLin t2 in
        {| Cstr.coefs:=Lin.add lt1 (Lin.neg lt2) ; Cstr.typ:=cmp; Cstr.cst:= (QNum.sub aux2 aux1) |}.
  Extraction Inline toCstrX.

  Definition toCstr (t1: QTerm.t) (cmp:cmpT) (t2:QTerm.t) : option Cstr.t :=
    match cmp with
      | Eq => Some (toCstrX t1 CstrC.Eq t2)
      | Le => Some (toCstrX t1 CstrC.Le t2)
      | Lt => Some (toCstrX t1 CstrC.Lt t2)
      | Neq => None
    end.

  Lemma SwitchSides1: forall (cmp:CstrC.cmpT) (n1 n2 n3: QNum.t),
    CstrC.cmpDenote cmp n1 (QNum.add n2 n3) <->
    CstrC.cmpDenote cmp (QNum.add n1 (QNum.neg n3)) n2.
  Proof.
    intros; destruct cmp; simpl; intuition subst; 
      try (ring 
        || (autorewrite with num in * |- *; auto)).
  Qed.

  Lemma SwitchSides2: forall (cmp: CstrC.cmpT) (n1 n2 n3: QNum.t),
    CstrC.cmpDenote cmp (QNum.add n2 n3) n1 <->
    CstrC.cmpDenote cmp n2 (QNum.add n1 (QNum.neg n3)).
  Proof.
    intros; rewrite SwitchSides1; split; intro h.
    - ring_simplify; auto.
    - ring_simplify in h; auto.
  Qed.

  Lemma toCstrX_correct t1 cmp t2 m:
    Cstr.sat (toCstrX t1 cmp t2) m <-> CstrC.cmpDenote cmp (QTerm.eval t1 m) (QTerm.eval t2 m).
  Proof.
    unfold toCstrX, Cstr.sat; simpl; rewrite !QTerm.toLin_correct.
    destruct (QTerm.toLin t1) as [lt1 aux1], (QTerm.toLin t2) as [lt2 aux2]; simpl.
    autorewrite with linterm num.
    rewrite <- SwitchSides1.
    rewrite (QNum.AddComm aux2 (QNum.neg aux1)).
    rewrite <- QNum.AddAssoc.
    rewrite (QNum.AddComm (QNum.neg _) _).
    rewrite SwitchSides2.
    rewrite (QNum.AddComm aux2 _).
    intuition.
  Qed.

  Lemma toCstr_correct t1 cmp t2: 
    WHEN c <- (toCstr t1 cmp t2) THEN
    forall m, Cstr.sat c m <-> cmpDenote cmp (QTerm.eval t1 m) (QTerm.eval t2 m).
  Proof.
    destruct cmp; simpl; intros; simpl; try (rewrite toCstrX_correct; simpl); intuition.
  Qed.

  Hint Resolve QCond.toCstr_correct: vpl.


End QCond.

Open Scope impure.

(* a tiny functor that lift conditions from "Cstr" to "QCond" *)
Module QCondAssume (D: BasicDomain QNum) (CstrD: HasAssume QNum Cstr D) <: HasAssume QNum QCond D.

  Import QCond.
  Import D.

(** assume: 
  for good precision, "Not" and "Neq" must have been eliminated from "c" 
  (using for instance "nnf_for_assume above")
*)

  Fixpoint assume (c:QCond.t) (a: t):  imp t :=
    match c with
      | Basic b => pure (if b then a else bottom)
      | Atom cmp tl tr => 
        match QCond.toCstr tl cmp tr with
          | None => pure (failwith "PolLinQ.assume:Neq" top)
          | Some c' => CstrD.assume c' a
        end
      | BinL AND cl cr => BIND aux <- assume cl a -;
                          assume cr aux
      | BinL OR cl cr => BIND auxl <- assume cl a -;
                         BIND auxr <- assume cr a -;
                         join auxl auxr
      | Not c0 => pure (failwith "PolLinQ.assume:Not" top)
    end.

  Lemma assume_correct: forall c a, WHEN p <- assume c a THEN 
     forall m, sat c m -> gamma a m -> gamma p m.
  Proof.
    induction c; simpl.
    - destruct b; simpl; VPLAsimplify.
    - VPLAsimplify.
      simpl in * |- *. intros H0 m; rewrite <- H. auto.
    - case op; simpl; xasimplify ltac:(intuition eauto with vpl).
    - VPLAsimplify.
  Qed.

End QCondAssume.

(* a naive generic version deriving "require" from "assume" (using negation !) *)
Module NaiveRequire (N: NumSig) (Import Cond:AffCondSig N) (Import D: WeakAbstractDomain N Cond) <: HasRequire N Cond D.

  Open Scope impure.

  Definition require (c:cond) (a:t) := 
    BIND aux <- assume (nnfForAssume (Not c)) a -;
    trace DEBUG "@naiveRequire@" (isBottom aux).

  Lemma require_correct (c:cond) (a:t): If require c a THEN forall m, (gamma a m) -> (sat c m).
  Proof.
    unfold require; unfold trace; VPLAsimplify.
    intros m; case (sat_dec c m); auto.
    intros H2 H3;  case (H0 m). 
    auto with vpl.
  Qed.
  Hint Resolve require_correct: vpl.
  Global Opaque require.

  Close Scope impure.

End NaiveRequire.



Module Type HasAtomicRequire (N: NumSig) (Import Cond: AffCondSig N) (Import D: BasicDomain N).
  
  Parameter atomicRequire: cmpT -> Term.t -> Term.t -> t -> imp (option bool).
  Parameter atomicRequire_correct: 
    forall t1 t2 cmp a, 
    WHEN o <- atomicRequire cmp t1 t2 a THEN
    (WHEN b <- o THEN 
     (If b THEN forall m, gamma a m -> sat (Atom cmp t1 t2) m))%option.
  Hint Resolve atomicRequire_correct: vpl.

End HasAtomicRequire.


(* an atomic require on "QCond" (from a require on Cstr) *)
Module QCondHasAtomicRequire (Import D: BasicDomain QNum) (Import CR: HasRequire QNum Cstr D)
  <: HasAtomicRequire QNum QCond D.

  Hint Resolve require_correct QCond.toCstr_correct: polC.

  Import String.
  Local Open Scope string_scope.

  Definition atomicRequire (cmp:cmpT) (tl tr:QTerm.t) (a:t) : imp (option bool) :=
    SOME c <- QCond.toCstr tl cmp tr -;
    let c:=trace DEBUG ("[atomicRequire " ++ (Cstr.pr c) ++ "]") c in
    BIND b <- require c a -;
    pure (Some b).

  Lemma atomicRequire_correct: 
    forall t1 t2 cmp a, 
      WHEN o <- atomicRequire cmp t1 t2 a THEN 
       (WHEN b <- o THEN 
         If b THEN forall m, gamma a m -> QCond.sat (QCond.Atom cmp t1 t2) m)%option.
  Proof.
    unfold atomicRequire, trace; xasimplify ltac:(eauto with polC). simpl in * |- *.
    intros; rewrite <- H. auto.
  Qed.

  Global Opaque atomicRequire.

End QCondHasAtomicRequire.


(* a generic require from an atomic one (and a assume) *)
Module AffCondRequire (N: NumSig) 
  (Import Cond:AffCondSig N) 
  (Import D:WeakAbstractDomain N Cond)  
  (Import AR: HasAtomicRequire N Cond D)
  <: HasRequire N Cond D.

  Module Naive := NaiveRequire N Cond D.

  Fixpoint require (c:Cond.t) (a: t):  imp bool :=
    match c with
      | Basic b => if b then pure true else isBottom a
      | Atom cmp tl tr => 
        BIND aux <- atomicRequire cmp tl tr a -;
        match aux with
          | None => Naive.require c a
          | Some b => pure b
        end        
      | BinL AND cl cr => 
        BIND b1 <- require cl a -;
        if b1 
        then require cr a 
        else pure false 
      | c0 => Naive.require c0 a
    end.

(* NOTE on "require" implementation:

By using "&&&" for AND, we do not loose precision.

This does not work for "OR" because
  x <= 2  implies  "x <= 1" OR "1 <= x <= 2"
But it does not implies neither of these alternatives !


But, by going into negation:

 assume "x > 1 /\ ( x < 1 \/ x > 2)" in "x <= 2"

reduces to "bottom" with the algorithm of QCondAssume !

And,
 
  assume "( x < 1 \/ x > 2) /\ x > 1" in "x <= 2"


also reduces to "bottom" !
*)

  Lemma require_correct (c:cond) (a:t): If require c a THEN forall m, (gamma a m) -> (sat c m).
  Proof.
    induction c; try (VPLAsimplify; fail).
    - case b; VPLAsimplify.
    - (* AND *) case op; simpl; VPLAsimplify.
  Qed.

End AffCondRequire.

Close Scope impure.

