(** This module implements the basic abstract domain of polyhedra
on affine conditions over rationals ([AffCond.QCond]).

It builds on [ConsSet] and adds to it the coupling with opaque
Ocaml polyhedra, operations on which serve as an oracle for
operations in [ConsSet]. It also adds the possibility for a
polyhedron to be empty. 

*)

Require Import List.
Require Import String.
Require CoqAddOn.
Require Import LinTerm.
Require Import ConsSet.
Require Shadow.
Require Import Itv.
Require Export AffCond.
Require Import Debugging.
Require Import AssignD.
Require Export Impure.

Import QCond.
Open Scope impure.

Module BasicPolQ <: BasicDomain QNum.

(** Coupling of a [ConsSet] with an opaque Ocaml polyhedron. *)
  Record polT: Type := mk {
    cons: Cs.t;
    ml: Shadow.t
  }.

  Definition polPr: polT -> string
    := fun p =>
      CoqAddOn.sprintf "{coq:\n%s\nocaml:\n%s}"
      ((Cs.pr (cons p))::(Shadow.pr (ml p))::nil).

  Definition polTop: polT
    := mk Cs.top Shadow.top.

(** Empty polyhedron is encoded as [None], 
    [Some p] encodes a non-empty polyhedron
*)
  Definition t:=option polT.

  Local Open Scope string_scope.
  Definition pr: t -> string
    := fun p =>
      match p with
        | None => "bot"
        | Some p' => "notbot " ++ (polPr p')
      end.


  Definition gamma (a: t) (m: Mem.t QNum.t) :=
    (EXISTS p <- a SUCH Cs.sat (cons p) m)%option.

   Instance gamma_ext: Proper (Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> iff) gamma.
   Proof.
     eapply sat_compat_iff; unfold gamma.
     intros x y H m1 m2 H0 H1. subst; PolCsimplify. 
     erewrite (eval_pointwise_compat Cs.sat_mdo); eauto.
   Qed.

  Definition top: t
    := Some polTop.

  Lemma top_correct: forall m, gamma top m.
    simpl; auto.
  Qed.
  Hint Resolve top_correct: vpl.

  Definition bottom: t := None.

  Lemma bottom_correct: forall m, ~(gamma bottom m).
  Proof.
    simpl; auto.
  Qed.

  Definition isIncl (p1 p2:t) : imp bool
    :=  match p1 with
          | None => pure true
          | Some pol1 =>
            match p2 with
              | None => 
                BIND isEmpty <- Shadow.isEmpty (ml pol1) -;
                pure match isEmpty with
                     | None => false
                     | Some cert => Cs.chkEmpty (cons pol1) cert
                     end
              | Some pol2 =>
                BIND isIncl <- Shadow.isIncl (ml pol1, ml pol2) -;
                pure match isIncl with
                     | None => false
                     | Some cert => Cs.chkIncl (cons pol1) (cons pol2) cert
                     end
            end
        end.

  Lemma isIncl_correct p1 p2: If isIncl p1 p2 THEN forall m, gamma p1 m -> gamma p2 m.
  Proof.
    unfold isIncl. xasimplify ltac:(eauto with polC).
  Qed.
  Hint Resolve isIncl_correct: vpl.
  Global Opaque isIncl.


  Definition meet (p1 p2:t): imp t :=
    SOME p1' <- p1 -;
    SOME p2' <- p2 -;
    BIND meet <- Shadow.meet (ml p1', ml p2') -;
    pure ( 
      let (maybeP, ce) := meet in
      match maybeP with
        | Some p => Some (mk (Cs.meet (cons p1') (cons p2') ce) p) 
        | None =>
          if Cs.meetEmpty (cons p1') (cons p2') ce
            then None
            else failwith "PolC.meet" top
      end).

  Lemma meet_correct p1 p2: WHEN p <- meet p1 p2 THEN forall m, gamma p1 m -> gamma p2 m ->gamma p m.
  Proof.
    unfold meet. xasimplify ltac:(eauto with polC).
  Qed.

  Definition join (p1 p2:t) : imp t := 
    match p1, p2 with
      | None, None => pure None
      | None, pol
      | pol, None => pure pol
      | Some pol1, Some pol2 =>
        BIND join <- Shadow.join (ml pol1, ml pol2) -;
        let (shadow, cert) := join in
          pure (Some (mk (Cs.buildJoin (cons pol1) (cons pol2) cert) shadow))
    end.

  Lemma join_correct p1 p2: WHEN p <- join p1 p2 THEN forall m, gamma p1 m \/ gamma p2 m -> gamma p m.
  Proof.
    unfold join; xasimplify ltac:(intuition eauto with polC).
  Qed.
  Hint Resolve join_correct: vpl.

  Definition widen (p1 p2: t): imp t :=
    match p1, p2 with
      | None, None => pure None
      | None, p | p, None => pure p
      | Some pol1, Some pol2 =>
        BIND widen <- Shadow.widen (ml pol1, ml pol2) -;
        let (cs, shadow) := widen in
          pure (Some (mk shadow cs))
    end.

  Definition project (p: t) (x: PVar.t): imp t
    := SOME p1 <- p -;
       BIND project <- Shadow.project (ml p1, x) -;
       pure (
         let (p2, c) := project in
         let cs:= Cs.approx (cons p1) c in
           if Cs.isFree x cs then
             Some (mk cs p2)
           else
             failwith ("PolC.project: "++(PVar.pr x)++" not free in "++(Cs.pr cs)) top).

  Lemma project_correct a x: WHEN p <- project a x THEN forall m v, gamma a m -> gamma p (Mem.assign x v m).
  Proof.
    unfold project; xasimplify ltac:(eauto with polC).
    intros; erewrite (mdoExt_free Cs.sat_mdo); eauto with polC progvar.
  Qed.

End BasicPolQ.

Module PolQItv <: HasGetItv QNum QTerm QItv BasicPolQ.

  Import BasicPolQ.

  Definition buildLow (p: polT) (v: Lin.t) (sItv: Shadow.itvT): QItv.bndT :=
      match Shadow.low sItv with
        | Shadow.Infty => QItv.Infty
        | Shadow.Open n f =>
          let c:=Cs.combine (cons p) f in
            if Cstr.isIncl c (Cstr.lowerToCstr v n) then QItv.Open n else QItv.Infty
        | Shadow.Closed n f =>
          let c:=Cs.combine (cons p) f in
            if Cstr.isIncl c (Cstr.lowerOrEqualsToCstr v n) then QItv.Closed n else QItv.Infty
      end.

  Lemma buildLow_correct (p: polT) (v: Lin.t) (sItv: Shadow.itvT) m: 
    gamma (Some p) m -> QItv.satLower (buildLow p v sItv) (Lin.eval v m).
  Proof.
    unfold buildLow; destruct (Shadow.low sItv); simpl; auto; PolCsimplify.
  Qed.

  Definition buildUp (p: polT) (v: Lin.t) (sItv: Shadow.itvT): QItv.bndT :=
      match Shadow.up sItv with
        | Shadow.Infty => QItv.Infty
        | Shadow.Open n f =>
          let c:=Cs.combine (cons p) f in
            if Cstr.isIncl c (Cstr.upperToCstr v n) then QItv.Open n else QItv.Infty
        | Shadow.Closed n f =>
          let c:=Cs.combine (cons p) f in
            if Cstr.isIncl c (Cstr.upperOrEqualsToCstr v n) then QItv.Closed n else QItv.Infty
      end.

  Lemma buildUp_correct (p: polT) (v: Lin.t) (sItv: Shadow.itvT) m: 
    gamma (Some p) m -> QItv.satUpper (buildUp p v sItv) (Lin.eval v m).
  Proof.
    unfold buildUp; destruct (Shadow.up sItv); simpl; auto; PolCsimplify.
  Qed.

  (** wrapping "mk" function of QItv *)
  Definition QItv_mk (l u: QItv.bndT): QItv.t
   := projT1 (QItv.mk l u).

  Lemma QItv_mk_correct l u x: QItv.satLower l x -> QItv.satUpper u x -> QItv.sat (QItv_mk l u) x.
  Proof.
    unfold QItv_mk. destruct (QItv.mk l u); simpl; auto.
  Qed.

  Definition getItv (p: t) (v: Lin.t): imp QItv.t := 
    match p with
    | None => pure QItv.Bot
    | Some pol =>
      BIND sItv <- Shadow.getItv (ml pol, v) -;
      let low := buildLow pol v sItv in
      let up := buildUp pol v sItv in
          pure (QItv_mk low up)
    end.

  Local Hint Resolve buildUp_correct buildLow_correct  QItv_mk_correct: polC.

  Lemma getItv_correct (p: t) (v: Lin.t):
    WHEN i <- getItv p v THEN forall m, gamma p m -> QItv.sat i (Lin.eval v m).
  Proof.
    unfold getItv; xasimplify ltac: (eauto with polC).
  Qed.

  Local Hint Resolve getItv_correct: polC.

  Definition get_itv: QTerm.t -> t -> imp QItv.t
    := fun a ab =>
      let (t,n) := QTerm.toLin a in
      BIND i <- getItv ab t -;
      pure (QItv.shift i n).

  Lemma get_itv_correct: forall (a: QTerm.t) (ab: t),
    WHEN i <- get_itv a ab THEN forall m, gamma ab m -> QItv.sat i (QTerm.eval a m).
  Proof.
    intros a ab; unfold get_itv.
    xasimplify ltac: (eauto with polC). simpl in * |- *.
    intros; rewrite QTerm.toLin_correct.
    rewrite QNum.AddComm.
    apply QItv.ShiftsatSpec.
    auto with polC.
  Qed.

End PolQItv.


Module CstrPol <: HasConcreteConditions QNum Cstr BasicPolQ.

  Import BasicPolQ.

  Definition assume (c: Cstr.t) (p: t): imp t :=
    SOME p1 <- p -;
    BIND cid <- Shadow.freshId (ml p1) -;
    BIND aux <- Shadow.add (ml p1, c) -;
    pure (let (opt, cert) := aux in
      match opt with
        | None =>
          if (Cs.chkEmpty (Cs.add (cons p1) cid c) cert) then 
            None
          else
            failwith "CstrPol.assume" top
        | Some s => Some (mk (Cs.redAdd (cons p1) cid c cert) s)
      end).
  
  Lemma assume_correct: forall c a, WHEN p <- assume c a THEN forall m, Cstr.sat c m -> gamma a m -> gamma p m.
  Proof.
    unfold assume; xasimplify ltac:(unfold not; eauto with polC).
  Qed. 

(** [fromCstr c] builds a polyhedron from a constraint.
  If it fails, [default] is returned. 
  Hence [default] may be chosen according to the fact that precision or soundness is required *)
Definition fromCstr: Cstr.t -> t -> imp t
    := fun c default =>
         BIND cid <- Shadow.freshId Shadow.top -;
         BIND res <- Shadow.addp (Shadow.top,c) -;
         pure match res with
           | Shadow.Added p (fwd, bwd) =>
             match Cs.addp Cs.top cid c fwd bwd with
               | Some cs => Some (mk cs p)
               | None => failwith "CstrPol.fromCstr" default
             end
           | Shadow.Contrad ce =>
             if Cs.chkEmpty (Cs.add Cs.top cid c) ce
             then None
             else failwith "CstrPol.fromCstr" default
         end.

(** Currently useless ??  TODO: adapt with [default] 
  Lemma fromCstr_correct: forall c p, fromCstr c = Some p ->
                                      forall m, Cstr.sat c m -> gamma p m.
  Proof.
    intros c p hr m hc.
    unfold fromCstr in hr.
    destruct (Shadow.addp Shadow.top c) as [p' [fwd bwd]|ce].
    - assert (pf := fun cs' =>
                      Cs.addp_correct Cs.top cs' (Shadow.getId Shadow.top)
                                      c fwd bwd m).
      destruct (Cs.addp Cs.top (Shadow.getId Shadow.top) c fwd bwd) as [cs|].
      + inversion hr.
        apply pf; [reflexivity|constructor|assumption].
      + discriminate hr.
    - assert (pf := Cs.chkEmpty_correct
                      (Cs.add Cs.top (Shadow.getId Shadow.top) c) ce).
      destruct (Cs.chkEmpty (Cs.add Cs.top (Shadow.getId Shadow.top) c) ce).
      + inversion hr.
        contradiction pf with (m := m).
        apply Cs.AddsatSpec; [constructor|assumption].
      + discriminate hr.
  Qed.
*)
  Local Hint Resolve Cs.addp_precise_c: polC.

  Lemma fromCstr_precise c: WHEN p <- fromCstr c None THEN forall m, gamma p m -> Cstr.sat c m.
  Proof.
    unfold fromCstr; unfoldDebug; xtsimplify ltac:idtac; intros m.
    destruct exta0 as [p' [fwd bwd]|ce].
    - apply try_catch_decomp.
      PolCsimplify.
    - apply if_decomp.
      PolCsimplify.
  Qed.

(** Test whether a polyhedron is included in a given half space. *)
  Definition require (c: Cstr.t) (p: t): imp bool :=
    BIND p' <- fromCstr c None -;
    isIncl p p'.

  Local Opaque fromCstr.
  Local Hint Resolve fromCstr_precise isIncl_correct: polC.

  Lemma require_correct: forall c a, If require c a THEN forall m, gamma a m -> Cstr.sat c m.
  Proof.
    intros c a.
    unfold require.
    xasimplify ltac: (eauto with polC).
  Qed.

End CstrPol.


(* a naive (and inefficient) implementation of rename *)
Module NaiveRename (Import D: HasAssume QNum QCond BasicPolQ) <: HasRename QNum BasicPolQ.

  Import BasicPolQ.

  Definition rename (x y: PVar.t) (a:t): imp t :=
    BIND p <- assume (Eq (Term.Var y) (Term.Var x)) a -;
    project p x.

  Lemma rename_correct: forall x y a, WHEN p <- (rename x y a) THEN
    forall m, gamma a (Mem.assign x (m y) m) -> gamma p m.
  Proof.
    unfold Basics.impl, rename.
    intros; xasimplify ltac:(eauto with vpl). simpl in * |- *.
    intros m H0. eapply gamma_ext.
    Focus 3. 
      eapply project_correct; [ eauto | eapply H; [idtac | eauto]].
      autorewrite with progvar.
      unfold Mem.assign. case (PVar.eq_dec x y); auto.
    auto.
    autorewrite with progvar.
    intros x'; erewrite @Mem.assign_id; auto.
  Qed. 

End NaiveRename.

(* implementation of rename *)
Module Rename (Import D: HasAssume QNum QCond BasicPolQ) <: HasRename QNum BasicPolQ.

  Import BasicPolQ.

  Definition rename (x y: PVar.t) (a:t) : imp t :=
  SOME p <- a -;
  BIND aux <- Shadow.rename (x,y,ml p) -;
  pure (Some {| cons := Cs.rename x y (cons p) ; ml := aux |}).

  Lemma rename_correct: forall x y a, WHEN p <- (rename x y a) THEN
    forall m, gamma a (Mem.assign x (m y) m) -> gamma p m.
  Proof.
    unfold Basics.impl, rename.
    intros; xasimplify ltac:idtac.
    intros; rewrite Cs.rename_correct.
    auto.
  Qed. 

End Rename.



(* Gluing all this together *)
Module PolLinQ <: AbstractDomain QNum QCond 
  := BasicPolQ <+ AddIsBottom QNum <+ QCondAssume BasicPolQ CstrPol
  <+ QCondHasAtomicRequire BasicPolQ CstrPol <+ AffCondRequire QNum QCond
  <+ PolQItv
  <+ Rename.

Module FullPolLinQ <: FullItvAbstractDomain QNum QCond QItv 
  := MakeAssign QNum QCond PolLinQ PolLinQ <+ AssignLiftItv QNum QCond QItv PolLinQ PolLinQ PolQItv
                                           <+ AssignSimplePrinting QNum QCond PolLinQ PolLinQ PolLinQ
  .
