(* EqSet *)
let x = Var.fromInt 1
let y = Var.fromInt 2
let z = Var.fromInt 3
let t = Var.fromInt 4
let nxt = 5

let varPr: Var.t -> string
= fun _x ->
	let vars: (Var.t * string) list
	= [x, "x"; y, "y"; z, "z"; t, "t"]
	in
	try
		List.assoc _x vars
	with
	| Not_found -> "v" ^ (Var.pr _x)

let mkc t v c =
	Cstr.mk t (List.map (fun (n, x) -> (Nb.mk1 n, x)) v) (Nb.mk1 c)

let eq = mkc Cstr.Eq
let le = mkc Cstr.Le
let lt = mkc Cstr.Lt

let mask l =
	List.fold_left (fun m x -> Rtree.set None m x (Some x)) Rtree.Nil l

(* XXX: This function does not do any check on the inputs. *)
let mk: (Var.t * Cstr.t) list -> EqSet.t
=	let add: int * EqSet.t -> Var.t * Cstr.t -> int * EqSet.t
	= fun (i, s) (x, c) -> (i + 1, (x, Cons.mk i c)::s)
	in
	fun l -> Pervasives.snd (List.fold_left add (1, EqSet.nil) l)

let mks: (int * Cstr.t) list -> EqSet.t
= fun l ->
	match EqSet.addM EqSet.nil l with
	| EqSet.Added (s, _) -> s
	| EqSet.Bot _ -> invalid_arg "EqSet_t.mks"

let mkl: Cstr.t list -> EqSet.t
= fun l ->
	let (_, l') = List.fold_left (fun (i, l) c -> (i+1, (i, c)::l)) (0, []) l in
	mks (List.rev l')

let optxpr =
	function
	| None -> "None"
	| Some x -> "Some " ^ (varPr x)

let prProp = function
| Cstr.Trivial -> "trivial"
| Cstr.Contrad -> "contrad"
| Cstr.Nothing -> "nothing"

let isEqC: Cert.cstr_t -> Cert.cstr_t -> bool
= fun ce1 ce2 ->
	match ce1, ce2 with
	| Cert.Direct f1, Cert.Direct f2 -> Cert.isEqF f1 f2
	| Cert.SplitEq (f1, f2), Cert.SplitEq (f3, f4) -> Cert.isEqF f1 f3 && Cert.isEqF f2 f4
	| Cert.Direct _, Cert.SplitEq (_, _)
	| Cert.SplitEq (_, _), Cert.Direct _ -> false

let impliesEq: (int * Cert.cstr_t) list -> (int * Cert.cstr_t) list -> bool
= fun l1 l2 ->
	let mEq (i1, cert1) (i2, cert2) = i1 = i2 && isEqC cert1 cert2 in
	try
		List.for_all2 mEq l1 l2
	with
	| Invalid_argument _ -> false

let impliesPr: (int * Cert.cstr_t) list -> string
= fun l ->
	let pr (i, cert) = Printf.sprintf "%i: %s" i (Cert.prC cert) in
	String.concat "\n" (List.map pr l)

let relEq: EqSet.rel_t -> EqSet.rel_t -> bool
= fun r1 r2 ->
	match r1, r2 with
	| EqSet.NoIncl, EqSet.NoIncl -> true
	| EqSet.Incl l1, EqSet.Incl l2 -> impliesEq l1 l2
	| EqSet.NoIncl, EqSet.Incl _
	| EqSet.Incl _, EqSet.NoIncl -> false

let relPr: EqSet.rel_t -> string
= function
	| EqSet.NoIncl -> "NoIncl"
	| EqSet.Incl l -> Printf.sprintf "Incl l with l =\n%s" (impliesPr l)

(* EqSet.filter *)
let filterTs: T.testT
=
	let chkNo (name, _, c, s, _) = fun state ->
		let c1 = EqSet.filter s c in
		let vars = List.map (fun (x, _) -> x) s in
		let zcoef x = Nb.cmpz (Vec.get (Cstr.get_v (Cons.get_c c1)) x) = 0 in
		if List.for_all zcoef vars then
			T.succeed state
		else
			let err = Printf.sprintf "some variable has non-zero coefficient in %s\n" (Cons.prExt varPr c1) in
			T.fail name err state
	in
	let chkId (name, _, c, s, _) = fun state ->
		let c1 = EqSet.filter s c in
		if Cons.get_id c = Cons.get_id c1 then
			T.succeed state
		else
			let err = Printf.sprintf "c.id: %i, c1.id: %i" (Cons.get_id c) (Cons.get_id c1) in
			T.fail name err state
	in
	let chkProp (name, p, c, s, _) = fun state ->
		let c1 = EqSet.filter s c in
		let p1 = Cstr.tellProp (Cons.get_c c1) in
		if p = p1 then
			T.succeed state
		else
			let e = (prProp p) ^ " != " ^ (prProp p1) in
			T.fail name e state
	in
	let chkLin (name, _, c, s, cert) = fun state ->
		let c1 = EqSet.filter s (Cons.initc c) in
		let (_, actualCert) = Cons.extractc c1 in
		if isEqC actualCert cert then
			T.succeed state
		else
			let err = Printf.sprintf "expected: %s\nbut got: %s" (Cert.prC cert) (Cert.prC actualCert) in
			T.fail name err state
	in
	let tcs = [
		"nil0", Cstr.Nothing, Cons.mk 0 (eq [1, x] 0), mkl [], Cert.Direct (Cert.mkF [0, Nb.u]);
		"nil1", Cstr.Nothing, Cons.mk 1 (eq [1, x] 1), mkl [], Cert.Direct (Cert.mkF [1, Nb.u]);
		"nil2", Cstr.Nothing, Cons.mk 2 (eq [1, x; 2, y] 0), mkl [], Cert.Direct (Cert.mkF [2, Nb.u]);
		"one0", Cstr.Nothing, Cons.mk 3 (eq [2, x; 3, y] 1), mkl [
			eq [1, y] 1 ], Cert.Direct (Cert.mkF [3, Nb.mk 2 (-1); 0, Nb.mk 2 3]);

		"incl0", Cstr.Trivial, Cons.mk 4 (eq [2, x; 3, y] 1), mkl [
			eq [1, x; 1, y] 1;
			eq [1, y] (-1) ], Cert.Direct (Cert.mkF [4, Nb.u; 0, Nb.mk1 (-2); 1, Nb.negU]);

		"incl1", Cstr.Trivial, Cons.mk 4 (le [2, x; 3, y] 1), mkl [
			eq [1, x; 1, y] 1;
			eq [1, y] (-1) ], Cert.Direct (Cert.mkF [4, Nb.u; 0, Nb.mk1 (-2); 1, Nb.negU]);

		"triv0", Cstr.Trivial, Cons.mk 4 (le [] 1), mkl [
			eq [1, x] 1;
			eq [1, y] 2 ], Cert.Direct (Cert.mkF [4, Nb.u]);

		"contrad0", Cstr.Contrad, Cons.mk 5 (eq [2, x; 1, z] 1), mkl [
			eq [1, x; 1, z] 0;
			eq [1, x] 2 ], Cert.Direct (Cert.mkF [5, Nb.u; 0, Nb.negU; 1, Nb.negU])
	] in
	T.suite "filter" [
		T.suite "no" (List.map chkNo tcs);
		T.suite "id" (List.map chkId tcs);
		T.suite "prop" (List.map chkProp tcs);
		T.suite "lin" (List.map chkLin tcs)
	]

(* EqSet.addM *)
(* XXX: check that all is fine when adding multiple equalities *)
let addMTs: T.testT
=	let chk (nm, s, l, r)
	= fun st -> try
		let ar = EqSet.addM s l in
		if EqSet.meetEq r ar
		then T.succeed st
		else
			let estr = Printf.sprintf "expected:\n%s\ngot:\n%s\n"
				(EqSet.meetPr varPr r) (EqSet.meetPr varPr ar)
			in
			T.fail nm estr st
		with Failure "EqSet.addM" -> T.fail nm "caught Failure exception" st
	in
	let tcs: (string * EqSet.t * (int * Cstr.t) list * EqSet.meetT) list
	= [
		"concat0", EqSet.nil, [],
			EqSet.Added (EqSet.nil, ([], []));

		"concat1", mks [], [0, eq [1, x] 0], EqSet.Added (
			mks [0, eq [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u])],
			[0, Cert.Direct (Cert.mkF [0, Nb.u])]));

		"concat2", mks [0, eq [1, x] 0], [], EqSet.Added (
			mks [0, eq [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u])],
			[0, Cert.Direct (Cert.mkF [0, Nb.u])]));

		"concat3", mks [0, eq [1, y] 1],
			[1, eq [1, x] 0], EqSet.Added (mks [
				0, eq [1, y] 1;
				1, eq [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])],

			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])]));

		"adj0", mks [], [0, eq [2, x] 0], EqSet.Added (
			mks [0, eq [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.mk 2 1])],
			[0, Cert.Direct (Cert.mkF [0, Nb.mk1 2])]));

		"adj1", mks [], [0, eq [-1, x] 0], EqSet.Added (
			mks [0, eq [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.negU])],
			[0, Cert.Direct (Cert.mkF [0, Nb.negU])]));

		"comb0", mks [0, eq [1, x] 0],
			[1, eq [1, x; 1, y] 0], EqSet.Added (mks [
				0, eq [1, x] 0;
				1, eq [1, y] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [0, Nb.negU; 1, Nb.u])],

			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.u])]));

		"comb+adj0", mks [0, eq [1, x] 0],
			[1, eq [1, x; 2, y] 0], EqSet.Added (mks [
				0, eq [1, x] 0;
				1, eq [1, y] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [0, Nb.mk 2 (-1); 1, Nb.mk 2 1])],

			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.mk1 2])]));

		"redundant0", mks [0, eq [1, x] 0],
			[1, eq [1, x] 0],
			EqSet.Added (mks [0, eq [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u])],

			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [0, Nb.u])]));

		"redundant1", mks [
			0, eq [1, x] 0;
			1, eq [1, y] 1],
			[2, eq [1, x; 2, y] 2],
			EqSet.Added (mks [
				0, eq [1, x] 0;
				1, eq [1, y] 1],
			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])],

			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u]);
			2, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.mk1 2])]));

		"contrad0", mks [],
			[1, eq [] 1],
			EqSet.Bot (Cert.mkF [1, Nb.u]);

		"contrad1", mks [0, eq [1, x] 0],
			[1, eq [1, x] 1],
			EqSet.Bot (Cert.mkF [1, Nb.negU; 0, Nb.u]);

		"contrad2", mks [
			0, eq [1, x] 0;
			1, eq [1, y] 0],
			[2, eq [1, x; 2, y] 1],
			EqSet.Bot (Cert.mkF [0, Nb.mk 2 (-1); 1, Nb.mk1 (-1);
						2, Nb.mk 2 1]);
	] in
	T.suite "addM" (List.map chk tcs)

(* EqSet.pick *)
let pickTs: T.testT
=
	let chk (name, msk, c, r) = fun state ->
		let r1 = EqSet.pick msk c in
		if r = r1 then
			T.succeed state
		else
			let e = "expected " ^ (optxpr r) ^ " but got " ^ (optxpr r1) in
			T.fail name e state
	in
	let tcs = [
		"nil0", mask [], Cons.mk 0 (eq [] 0), None;
		"nil1", mask [x], Cons.mk 0 (eq [] 0), None;
		"nil2", mask [y], Cons.mk 0 (eq [] 0), None;
		"nil3", mask [], Cons.mk 0 (eq [1, x] 0), None;
		"diff0", mask [x], Cons.mk 0 (eq [1, y] 0), None;
		"one0", mask [x], Cons.mk 0 (eq [1, x] 0), Some x;
		"one1", mask [x], Cons.mk 0 (eq [1, x; 1, y] 0), Some x;
		"one2", mask [y], Cons.mk 0 (eq [1, x; 1, y] 0), Some y;
		"m0", mask [x; y], Cons.mk 0 (eq [1, y] 0), Some y;
		"m1", mask [x; y], Cons.mk 0 (eq [1, y; 1, z] 0), Some y;

		(* implementation detail *)
		"m2", mask [x; y], Cons.mk 0 (eq [1, x; 1, y] 0), Some x
	] in
	T.suite "pick" (List.map chk tcs)

(* EqSet.subst *)
let substTs: T.testT
=
	let chkId (name, x, i, i1, c, s, _) = fun state ->
		let c1 = Cons.mk 0 c in
		let (i2, _) = EqSet.subst x i c1 s in
		if i1 = i2 then
			T.succeed state
		else
			let e =
				"expected " ^ (string_of_int i1) ^
				" but got " ^ (string_of_int i2)
			in
			T.fail name e state
	in
	let chkCons (name, x, i, _, c, s, s1) = fun state ->
		let c1 = Cons.mk 0 c in
		let (_, s2) = EqSet.subst x i c1 s in
		if EqSet.isEq s1 s2 then
			T.succeed state
		else
			let e =
				"expected:\n" ^  (EqSet.prExt varPr s1) ^ "but got:\n" ^ (EqSet.prExt varPr s2)
			in
			T.fail name e state
	in
	let tcs = [
		"nil0", x, 0, 0, eq [1, x] 0, mk [], mk [];
		"nil1", y, 0, 0, eq [1, x] 0, mk [], mk [];
		"nil2", y, 1, 1, eq [1, x] 0, mk [], mk [];
		"one0", x, 0, 1, eq [1, x] 0,
			mk [y, eq [1, x; 1, y] 1],
			mk [y, eq [1, y] 1];

		"one1", x, 1, 2, eq [1, x] 0,
			mk [y, eq [1, x; 1, y] 1],
			mk [y, eq [1, y] 1];

		"no0", x, 0, 1, eq [1, x] 0, mk [y, eq [1, y] 1], mk [y, eq [1, y] 1];
		"m0", x, 0, 2, eq [1, x] 0, mk [
			z, eq [1, z] 1;
			y, eq [1, x; 1, y] 1
		], mk [
			z, eq [1, z] 1;
			y, eq [1, y] 1 ];
	] in
	T.suite "subst" [
		T.suite "id" (List.map chkId tcs);
		T.suite "cons" (List.map chkCons tcs)
	]

(* EqSet.tryDefs *)
let tryDefsTs: T.testT
=
	let chk_int (t, _, i0, msk, s) = fun state ->
		let (_, i1, s1) = EqSet.tryDefs i0 msk s in
		let a = List.for_all
			(fun c -> i0 <= (Cons.get_id c) && (Cons.get_id c) < i1) (EqSet.list s1)
		in
		if a then
			T.succeed state
		else
			let e =
				"not all ids between " ^ (string_of_int i0) ^
				" and " ^ (string_of_int i1) ^ "\n" ^ (EqSet.prExt varPr s1)
			in
			T.fail t e state
	in
	let chk_dup (t, _, i0, msk, s) = fun state ->
		let (_, _, s1) = EqSet.tryDefs i0 msk s in
		let ids = List.sort Pervasives.compare
			(List.map (fun c -> (Cons.get_id c)) (EqSet.list s1))
		in
		let rec chk i1 =
			function
			| [] -> true
			| i2::l ->
				if i1 = i2 then
					false
				else
					chk i2 l
		in
		let a =
			match ids with
			| [] -> true
			| i::l -> chk i l
		in
		if a then
			T.succeed state
		else
			let e = "found duplicate ids\n" ^ (EqSet.prExt varPr s1) in
			T.fail t e state
	in
	let chk_res (t, r, i0, msk, s) = fun state ->
		let (def, _, _) = EqSet.tryDefs i0 msk s in
		let a = if def = None then false else true in
		if a = r then
			T.succeed state
		else
			T.fail t "bad conclusion" state
	in
	let chk_def (t, _, i0, msk, s) = fun state ->
		let (def, _, _) = EqSet.tryDefs i0 msk s in
		match def with
		| None -> T.succeed state
		| Some (_, x) ->
			if List.exists (fun (x1, _) -> x = x1) s then
				T.succeed state
			else
				let e = "not found" in
				T.fail t e state
	in
	let chk_nox (t, _, i0, msk, s) = fun state ->
		let (def, _, s1) = EqSet.tryDefs i0 msk s in
		match def with
		| None -> T.succeed state
		| Some (_, x) ->
			let noXPred c = Nb.cmpz (Vec.get (Cstr.get_v (Cons.get_c c)) x) = 0 in
			if List.for_all (fun (_, c) -> noXPred c) s1 then
				T.succeed state
			else
				T.fail t "x remaining" state
	in
	let tcs = [
		"nil0", false, 0, mask [], mkl [];
		"nil1", false, 0, mask [x], mkl [];
		"one0", true, 0, mask [x], mkl [eq [1, x] 0];
		"one0", true, 0, mask [x], mkl [
			eq [1, x] 0;
			eq [1, x; 1, y] 0 ];

		"no0", false, 2, mask [z], mkl [
			eq [1, x; 1, z] 0;
			eq [1, y] 0 ]
	] in
	T.suite "tryDefs" [
		T.suite "int" (List.map chk_int tcs);
		T.suite "dup" (List.map chk_dup tcs);
		T.suite "res" (List.map chk_res tcs);
		T.suite "def" (List.map chk_def tcs);
		T.suite "nox" (List.map chk_nox tcs)
	]

(* EqSet.trySubstM *)
let trySubstMTs: T.testT
=
	let chk_var (t, i, x, msk, s) = fun state ->
		let (optx1, _, _) = EqSet.trySubstM i msk s in
		match optx1 with
		| None ->
			if x = None then
				T.succeed state
			else
				let e = "expected " ^ (optxpr x) ^ " but got None" in
				T.fail t e state
		| Some (_, x1) ->
			if x = Some x1 then
				T.succeed state
			else
				let e = "expected " ^ (optxpr x) ^ " but got " ^ (optxpr (Some x1)) in
				T.fail t e state
	in
	let chk_nox (t, i, x, msk, s) = fun state ->
		let (e, s1, s2) =
			let (_, i1, s1) = EqSet.trySubstM i msk s in
			match x with
			| None -> ("trySubstM did change s", s, s1)
			| Some x ->
				let c = Cons.mk (i1 + 1) (eq [1, x] 0) in
				let (_, s2) = EqSet.subst x i c s1 in
				let e =
					"still some " ^ (varPr x) ^ " in\n" ^ (EqSet.prExt varPr s1)
				in
				(e, s1, s2)
		in
		if EqSet.isEq s1 s2 then
			T.succeed state
		else
			T.fail t e state
	in
	let tcs = [
		"x0", 0, Some x, mask [x], mk [
			y, eq [1, x; 1, y] 0;
			x, eq [1, x] 0 ];

		"no0", 0, None, mask [z], mk [
			x, eq [1, x] 0 ]
	] in
	T.suite "trySubstM" [
		T.suite "var" (List.map chk_var tcs);
		T.suite "nox" (List.map chk_nox tcs)
	]

(* EqSet.joinSetup *)
let joinSetupTs: T.testT
=
	let alpha = Var.fromInt nxt in
	let nxt' = nxt + 1 in
	let chk_id0 dup (t, idOff, s) = fun state ->
		let (_, _, s1) = EqSet.joinSetup dup (Var.fromInt nxt') Rtree.Nil alpha idOff s in
		let sorted =
			let ids = List.map (fun c -> (Cons.get_id c)) (EqSet.list s) in
			List.sort Pervasives.compare ids
		in
		let sorted1 =
			let ids1 = List.map (fun c -> (Cons.get_id c) - idOff) (EqSet.list s1) in
			List.sort Pervasives.compare ids1
		in
		if sorted = sorted1 then
			T.succeed state
		else
			T.fail t "not equal" state
	in
	let chk_def0 dup (t, i0, s) = fun state ->
		let (_, reloc, s1) = EqSet.joinSetup dup (Var.fromInt nxt') Rtree.Nil alpha i0 s in
		let a =
			let chk (x1, _) (x2, _) =
				if dup then
					x1 = x2
				else
					Rtree.get None reloc x1 = Some x2
			in
			List.for_all2 chk s s1
		in
		if a then
			T.succeed state
		else
			let err =
				Printf.sprintf "wrong defined variables: s:\n%s\ns1:\n%s\n"
				(EqSet.prExt varPr s) (EqSet.prExt varPr s1)
			in
			T.fail t err state
	in
	let chk_frag0 dup (t, i0, s) = fun state ->
		let (_, _, s1) = EqSet.joinSetup dup (Var.fromInt nxt') Rtree.Nil alpha i0 s in
		let a =
			let chk c =
				match Cons.get_f c with
				| [] -> false
				| (id, coef)::[] -> id = (Cons.get_id c) && Nb.cmp Nb.u coef = 0
				| _ -> false
			in
			List.for_all chk (EqSet.list s1)
		in
		if a then
			T.succeed state
		else
			T.fail t "bad fragments" state
	in
	let tcs = [
		"one0", 10, mk [ x, eq [1, x] 1 ];
		"one1", 10, mk [ x, eq [1, x; 1, y] 1 ];
		"one2", 10, mk [ y, eq [1, x; 1, y] 1 ];
		"m0", 10, mk [
			x, eq [1, x; 1, y] 1;
			y, eq [2, y; 1, z] 1;
			z, eq [1, z; 2, t] 1 ]
	] in
	let chk_id = chk_id0 false in
	let chk_idd = chk_id0 true in
	let chk_def = chk_def0 false in
	let chk_frag = chk_frag0 false in
	let chk_defd = chk_def0 true in
	let chk_fragd = chk_frag0 true in
	T.suite "joinSetup" [
		T.suite "id" (List.map chk_id tcs);
		T.suite "idd" (List.map chk_idd tcs);
		T.suite "def" (List.map chk_def tcs);
		T.suite "defd" (List.map chk_defd tcs);
		T.suite "frag" (List.map chk_frag tcs);
		T.suite "fragd" (List.map chk_fragd tcs)
	]

(* EqSet.isIncl *)
let isInclTs: T.testT
=
	let chk (name, s1, s2, expected) = fun state ->
		let actual = EqSet.isIncl s1 s2 in
		if relEq actual expected then
			T.succeed state
		else
			T.fail name (relPr actual) state
	in
	let tcs = [
		"id0", mks [0, eq [1, x] 0 ], mks [0, eq [1, x] 0 ], EqSet.Incl [0, Cert.Direct (Cert.mkF [0, Nb.u])];
		"id1", mks [
			0, eq [1, y] 0;
			1, eq [1, x] 0
		], mks [0, eq [1, x] 0 ], EqSet.Incl [0, Cert.Direct (Cert.mkF [1, Nb.u])];

		"bin0", mks [
			0, eq [1, y] 0;
			1, eq [1, x] 0
		], mks [0, eq [1, x; 1, y] 0 ], EqSet.Incl [0, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.u])]
	] in
	T.suite "isIncl" (List.map chk tcs)

(* EqSet.rename *)
let renameTs: T.testT
=
	let x' = Var.fromInt nxt in
	let chkVar (t, fromX, toY, s0) = fun state ->
		let defVars s = List.map (fun (v, _) -> v) s in
		let s = EqSet.rename s0 fromX toY in
		if List.for_all (fun v -> v <> fromX) (defVars s) then
			T.succeed state
		else
			let e = (varPr fromX) ^ " remaining\n" ^ (EqSet.prExt varPr s) in
			T.fail t e state
	in
	let chkCoef (t, fromX, toY, s0) = fun state ->
		let aX s =
			let coef c x = Vec.get (Cstr.get_v (Cons.get_c c)) x in
			List.map (fun (_, c) -> coef c fromX) s
		in
		let s = EqSet.rename s0 fromX toY in
		if List.for_all (fun a -> Nb.cmpz a = 0) (aX s) then
			T.succeed state
		else
			let e = (varPr fromX) ^ " remaining\n" ^ (EqSet.prExt varPr s) in
			T.fail t e state
	in
	let tcs = [
		"one0", x', x, mkl [ eq [1, x'] 0 ];
		"one1", x', x, mkl [ eq [1, x'; 1, y] 0 ];
		"no0", x', x, mkl [ eq [1, y] 0 ];
		"m0", x', x ,mkl [
			eq [1, x'; 1, y] 1;
			eq [1, x'; 2, y] 2 ];
	] in
	T.suite "rename" [
		T.suite "var" (List.map chkVar tcs);
		T.suite "coef" (List.map chkCoef tcs)
	]

let ts: T.testT
= T.suite "EqSet" [filterTs; addMTs; pickTs; substTs; tryDefsTs; trySubstMTs; joinSetupTs; isInclTs; renameTs]
