let x = Var.fromInt 1
let y = Var.fromInt 2
let z = Var.fromInt 3

let nxt = Var.fromInt 4

let varPr: Var.t -> string
= fun _x ->
	let vars: (Var.t * string) list
	= [x, "x"; y, "y"; z, "z"]
	in
	try
		List.assoc _x vars
	with
	| Not_found -> "v" ^ (Var.pr _x)

let v: (int * Var.t) list -> Vec.t
= fun l -> Vec.mk (List.map (fun (i, v) -> (Nb.mk1 i, v)) l)

(* Vec.elim *)
let elimInFromOnlyTs: T.testT
=
	let chk (name, using, from) = fun state ->
		try
			let _ = Vec.elim y using from in
			T.fail name "Invalid_argument not raised" state
		with
		| Invalid_argument "Vec.elim" -> T.succeed state
		| _ -> T.fail name "unexpected exception raised" state
	in
	let tcs = [
		"simple0", v [1, x], v [1, x; 1, y]
	] in
	T.suite "inFromOnly" (List.map chk tcs)

let elimNotInFromTs: T.testT
=
	let chk (name, using, from) = fun state ->
		let r = Vec.elim y using from in
		if Vec.isEq from r then
			T.succeed state
		else
			T.fail name (Vec.pr varPr r) state
	in
	let tcs = [
		"simple0", v [1, x; 1, y; 1, z], v [1, x];
		"nowhere0", v [1, x; 1, z], v [1, x]
	] in
	T.suite "notInFrom" (List.map chk tcs)

let elimInBothTs: T.testT
=
	let chk (name, using, from, expected) = fun state ->
		let r = Vec.elim y using from in
		if Vec.isEq r expected then
			T.succeed state
		else
			T.fail name (Vec.pr varPr r) state
	in
	let tcs = [
		"simple0", v [1, x; 1, y], v [-1, y], v [1, x]
	] in
	T.suite "inBoth" (List.map chk tcs)

let elimTs: T.testT
= T.suite "elim" [elimInFromOnlyTs; elimNotInFromTs; elimInBothTs]

(* Vec.shift *)
let shiftTs: T.testT
=
	(* Check that the coefficients in [vec] are to be found at the right place in [shiftedVec]. *)
	let chkCoefs (name, vec, relocTbl) = fun state ->
		let (_, shiftedVec, relocTbl1) = Vec.shift nxt vec relocTbl in
		let rec chk vec1 tbl1 =
			match vec1, tbl1 with
			| Rtree.Nil, _ -> true
			| _, Rtree.Nil -> false
			| Rtree.Sub (l, n, r), Rtree.Sub (lTbl, reloc, rTbl) ->
				let node =
					if Nb.cmpz n = 0 then
						true
					else
						match reloc with
						| None -> false
						| Some x -> Nb.cmp n (Vec.get shiftedVec x) = 0
				in
				node && chk l lTbl && chk r rTbl
		in
		if chk vec relocTbl1 then
			T.succeed state
		else
			T.fail name "difference between coefficients" state
	in
	(* Check that no more than the expected coefficients are to be found in the result.
	XXX: this way of checking is maybe overkill and redundant with [chkCoefs]. *)
	let chkNil (name, v, r) = fun state ->
		let (_, sv, nr) = Vec.shift nxt v r in
		let rec build nv v1 r1 =
			match v1, r1 with
			| Rtree.Nil, _ -> nv
			| _, Rtree.Nil -> failwith "Vec_t.shift_ts"
			| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
				let nv1 =
					if Nb.cmpz n1 = 0 then
						nv
					else
						match n2 with
						| None -> failwith "Vec_t.shift_ts"
						| Some x -> Rtree.set false nv x true
				in
				build (build nv1 l1 l2) r1 r2
		in
		let rec chk v1 r1 =
			match v1, r1 with
			| Rtree.Nil, Rtree.Nil -> true
			| Rtree.Nil, _ | _, Rtree.Nil -> false
			| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
				let a =
					if Nb.cmpz n1 = 0 then
						not n2
					else
						n2
				in
				a && chk l1 l2 && chk r1 r2
		in
		if chk sv (build Rtree.Nil v nr) then
			T.succeed state
		else
			T.fail name "unexpected non-nul coefficient" state
	in
	(* "Syntactic" criteria on the relocation table:
		- once set, a node does not change
		- all variables are allocated before the returned "nxt" *)
	let chkReloc (name, v, r) = fun state ->
		let (nxt1, _, nr) = Vec.shift nxt v r in
		let rec chk r nr =
			match r, nr with
			| Rtree.Nil, Rtree.Nil -> true
			| Rtree.Nil, Rtree.Sub (l, None, r) -> chk Rtree.Nil l && chk Rtree.Nil r
			| Rtree.Nil, Rtree.Sub (l, Some x, r) -> (Var.cmp x nxt1 = -1) && chk Rtree.Nil l && chk Rtree.Nil r
			| Rtree.Sub _, Rtree.Nil -> false
			| Rtree.Sub (_, Some _, _), Rtree.Sub (_, None, _) -> false
			| Rtree.Sub (l1, None, r1), Rtree.Sub (l2, None, r2) -> chk l1 l2 && chk r1 r2
			| Rtree.Sub (l1, None, r1), Rtree.Sub (l2, Some x, r2) -> (Var.cmp x nxt1 = -1) && chk l1 l2 && chk r1 r2
			| Rtree.Sub (l1, Some x1, r1), Rtree.Sub (l2, Some x2, r2) -> x1 = x2 && chk l1 l2 && chk r1 r2
		in
		if chk r nr then
			T.succeed state
		else
			T.fail name "bad relocation table" state
	in
	let x1 = nxt in
	let y1 = Var.next x1 in
	let tcs = [
		"nil0", Vec.mk [], Rtree.mk None [];
		"nil1", Vec.mk [], Rtree.mk None [x, Some x1];
		"one0", Vec.mk [Nb.u, x], Rtree.mk None [x, Some x1];
		"one1", Vec.mk [Nb.u, x], Rtree.mk None [];
		"one2", Vec.mk [Nb.mk1 2, x], Rtree.mk None [x, Some x1];
		"none0", Vec.mk [Nb.u, x; Nb.u, y], Rtree.mk None [];
		"some0", Vec.mk [Nb.u, x; Nb.u, y], Rtree.mk None [x, Some x1];
		"some1", Vec.mk [Nb.u, x; Nb.u, y], Rtree.mk None [y, Some y1];
		"all0", Vec.mk [Nb.u, x; Nb.u, y],
			Rtree.mk None [x, Some x1; y, Some y1]
	] in
	T.suite "shift" [
		T.suite "coefs" (List.map chkCoefs tcs);
		T.suite "nil" (List.map chkNil tcs);
		T.suite "reloc" (List.map chkReloc tcs)
	]

(* Vec.gcd *)
let gcdTs: T.testT
=
	let chk (name, r, v1) = fun state ->
		let r1 = Vec.gcd v1 in
		if Nb.cmp r r1 = 0 then
			T.succeed state
		else
			T.fail name (Nb.pr r1) state
	in
	let tcs = [
		"nil0", Nb.u, Vec.mk [];
		"u0", Nb.u, Vec.mk [Nb.u, x];
		"u1", (Nb.mk 2 1), Vec.mk [Nb.mk1 2, x];
		"m0", Nb.u, Vec.mk [Nb.u, x; Nb.mk1 2, y];
		"m1", Nb.mk 2 1, Vec.mk [Nb.mk1 2, x; Nb.mk1 2, y];
		"m2", Nb.mk 2 1, Vec.mk [Nb.mk1 2, x; Nb.mk1 4, y];
		"f0", Nb.mk1 2, Vec.mk [Nb.mk 2 1, x; Nb.mk 2 1, y];
		"f1", Nb.mk 3 2,
			Vec.mk [Nb.mk 2 9, x; Nb.mk 2 3, y];

		"hole0", Nb.mk 2 1, Vec.mk [Nb.mk1 4, y; Nb.mk1 6, z];
		"neg0", Nb.u, Vec.mk [Nb.negU, x]
	] in
	T.suite "gcd" (List.map chk tcs)

(* listVars *)
let listVarsTs: T.testT
=
	let isEq list1 list2 =
		if List.length list1 = List.length list2 then
			let sList1 = List.sort Var.cmp list1 in
			let sList2 = List.sort Var.cmp list2 in
			sList1 = sList2
		else
			false
	in
	let chk (name, vars, vecs) state =
		let vars1 = Vec.listVars vecs in
		if isEq vars vars1 then
			T.succeed state
		else
			T.fail name "not equal" state
	in
	let [x; y; z; t] = T.pathGen 4 in (* XXX: should be done globally. *)
	let tcs = [
		"empty0", [], [Vec.mk []];
		"empty1", [], [Vec.mk []; Vec.mk []];
		"one0", [x], [Vec.mk [Nb.u, x]];
		"one1", [t], [Vec.mk [Nb.u, t]];
		"one2", [z], [Vec.mk [Nb.u, z]];
		"m0", [x; y; z], [Vec.mk [Nb.u, x; Nb.mk1 2, z]; Vec.mk [Nb.u, y]]
	] in
	T.suite "listVars" (List.map chk tcs)

let ts: T.testT
= T.suite "Vec" [elimTs; shiftTs; gcdTs; listVarsTs]
